//
//cl generate_patches.c
//

/*
    search firmwares for hack addresses
    generates patches.h, patches.c for use by exTAP.c


    usage:
    e.g.
    to find all .raw files in current directory, use:
        generate_patches

    to find all .raw files in ..\..\firmware, use:
        generate_patches -p ..\..\firmware  -x raw

    to find all firmware_sys_456.bin in c:\, use:
        generate_patches -p c:\ -f firmware_sys_456 -x bin

    firmware name MUST contain the sysid in the form: <firmware_name>sys_456<more_firmware_name>

sysid = bytes 4,5 in .tfd file


    v0.01 - original
    v0.02 - removed fwlist_tap_load_data1
          - added fwlist_enable_tshift
          - added fwlist_stop_tshift
          - added fwlist_mheg_disable

*/
#ifdef _WIN32
#include <windows.h>
#include <stdio.h>
#include <io.h>
#include <wininet.h>
#include <time.h>
#include <string.h>

typedef ULONGLONG ulong64;
typedef ULONG     dword;
typedef USHORT    word;
typedef UCHAR     byte;
#define to_be(d) (((d&0xff)<<8)|((d&0xff00)>>8))

//-----------------------------------------------------------------------------
/*
    search for several types of item

    - function start addresses
    - static variables
    - $gp variables


*/
#define MAX_FOUND 0x10
//
// patt
// return: 0 = found, -VE = error
//
int search_pattern(byte *pattern, int pattern_length, byte *base, dword start, dword end, dword *result)
{
    int match;
    int found;
    int i,j,k;

    if (pattern_length < 1) return -1;
//    if ((start < 0x00000000) || (start > 0x00250000)) start = 0x00000000;
//    if ((end   < 0x00000000) || (end   > 0x00250000)) end   = 0x00250000;
    if ((start < 0x00000000) || (start > 0x001f6220)) start = 0x00000000;
    if ((end   < 0x00000000) || (end   > 0x001f6220)) end   = 0x001f6220;
    start &= 0xfffffffc;
    end &= 0xfffffffc;

//  printf("\nsearching from %08x to %08x for", start, end);
//  for (i=0; i<((pattern_length<0x10)?pattern_length:0x10); i++) {
//      printf(" %02x", pattern[i]);
//  }
//  printf("... ");

    found = 0;
    result[found] = (dword)start;
    for (i=0; i<(end-start-pattern_length); i+=4) {

//      printf("\n%08x: ", base+i);
//      for(k=0; k<8; k++) {
//          printf("%02x ", base[start+i+k]);
//      }
//      printf("  j=%x", j);

        result[found] = (dword)start+i;

        //opcodes not subject to wild
        //2nd field subject to wild (for jmps)
        //3rd,4th subject to wild (ff)
        for (j=0; j<pattern_length; j++) {
            if ((base[start+i+j] != pattern[j]) && (pattern[j] != 0xff)) {
                goto no_match;
            }
        }
match:
        found++;
        if (found >= MAX_FOUND) return found;
no_match:
        ;
    }
    //no match
    return found;
}








#define MAX_FIRMWARES 0x100
int num_firmwares;
char fw_filename[MAX_FIRMWARES][0x100];

typedef struct {
    dword fw_sysid[MAX_FIRMWARES];
    dword fw_appl_version[MAX_FIRMWARES];
    dword fw_gp[MAX_FIRMWARES];
    dword fw_tap_table[MAX_FIRMWARES];
    dword fw_tap_ident[MAX_FIRMWARES];
    dword fw_os_free[MAX_FIRMWARES];
    dword fw_add_task_queue[MAX_FIRMWARES];
    dword fw_shutdown_irq_disable[MAX_FIRMWARES];
    dword fw_shutdown_irq_enable[MAX_FIRMWARES];
    dword fw_shutdown_task[MAX_FIRMWARES];
    dword fw_process_tdt[MAX_FIRMWARES];
    dword fw_process_tot[MAX_FIRMWARES];
    dword fw_DbgSi_LoadDmx[MAX_FIRMWARES];
    dword fw_utc_time[MAX_FIRMWARES];
    dword fw_utctolocal_time[MAX_FIRMWARES];
    dword fw_process_eit[MAX_FIRMWARES];
    dword TAP_GetCurrentEvent[MAX_FIRMWARES];
    dword fw_event_table[MAX_FIRMWARES];
    dword fw_event_table_size[MAX_FIRMWARES];
    dword fw_max_events[MAX_FIRMWARES];
    dword fw_event_alloc[MAX_FIRMWARES];
    dword fw_event_buffer[MAX_FIRMWARES];
    dword fw_event_delete[MAX_FIRMWARES];
    dword fw_process_dsmcc[MAX_FIRMWARES];
    dword fw_mheg_entry[MAX_FIRMWARES];
    dword fw_hdd_move[MAX_FIRMWARES];
    dword fw_key_map_patch[MAX_FIRMWARES];
    dword fw_key_map_patch2[MAX_FIRMWARES];
    dword fw_led_patch[MAX_FIRMWARES];
    dword fw_display_func[MAX_FIRMWARES];
    dword fw_display_string[MAX_FIRMWARES];
    dword fw_ffindfirst[MAX_FIRMWARES];
    dword fw_tap_load[MAX_FIRMWARES];
    dword fw_tap_load_data0[MAX_FIRMWARES];
//    dword fw_tap_load_data1[MAX_FIRMWARES];
    dword fw_tap_load_data2[MAX_FIRMWARES];
    dword fw_num_ext_functions[MAX_FIRMWARES];
    dword fw_ext_function_table[MAX_FIRMWARES];

    dword fw_enable_tshift[MAX_FIRMWARES];
    dword fw_stop_tshift[MAX_FIRMWARES];
    dword fw_mheg_disable[MAX_FIRMWARES];
    dword fw_ioctl[MAX_FIRMWARES];
    dword fw_epgstck_patch[MAX_FIRMWARES];
    dword fw_channel_scale[MAX_FIRMWARES];
    dword fw_set_wss[MAX_FIRMWARES];
    dword fw_etc_structure[MAX_FIRMWARES];
    dword fw_afd[MAX_FIRMWARES];
    dword fw_sequence_header[MAX_FIRMWARES];
    dword fw_get_genre_str[MAX_FIRMWARES];
    dword fw_epg_highlighted_event[MAX_FIRMWARES];
    dword fw_epg_refresh[MAX_FIRMWARES];

}type_fw_addresses;


//fw_appl_version
byte pattern_fw_appl_version[] = {
    0x24, 0x19, 0xFF, 0xFF,                             //li      $t9, XXXX                 ; <---fw_appl_version
    0x27, 0x18, 0xFF, 0xFF,                             //la      $t8, TAP_SystemProc
    0xaf, 0xd9, 0x00, 0x2c,                             //sw      $t9, _appl_version($fp)
    0xaf, 0xd8, 0x00, 0x28,                             //sw      $t8, pTAP_SystemProc($fp)
    0x3c, 0x0f, 0xFF, 0xFF,                             //lui     $t7, XXXX
    0x3c, 0x19, 0xFF, 0xFF,                             //lui     $t9, XXXX
    0x3c, 0x18, 0xFF, 0xFF,                             //lui     $t8, XXXX
    0x25, 0xef, 0xFF, 0xFF,                             //la      $t7, TAP_GetState
    0x27, 0x39, 0xFF, 0xFF,                             //la      $t9, TAP_EnterNormal
    0x27, 0x18, 0xFF, 0xFF,                             //la      $t8, TAP_Exit
    0xaf, 0xcf, 0x00, 0x34,                             //sw      $t7, pTAP_GetState($fp)
    0xaf, 0xd9, 0x00, 0x3c,                             //sw      $t9, pTAP_EnterNormal($fp)
    0xaf, 0xd8, 0x00, 0x48,                             //sw      $t8, pTAP_Exit($fp)
    0x3c, 0x0e, 0xFF, 0xFF,                             //lui     $t6, XXXX
    0x3c, 0x0f, 0xFF, 0xFF,                             //lui     $t7, XXXX
    0x3c, 0x19, 0xFF, 0xFF,                             //lui     $t9, XXXX
};

//fw_appl_version
byte pattern_fw_appl_version2[] = {
    0x27, 0x18, 0xFF, 0xFF,                             //la      $t8, TAP_SystemProc
    0xac, 0x98, 0x00, 0x28,                             //sw      $t8, 0x28($a0)
    0x3c, 0x0f, 0xFF, 0xFF,                             //lui     $t7, XXXX
    0x3c, 0x0e, 0xFF, 0xFF,                             //lui     $t6, XXXX
    0x3c, 0x18, 0xFF, 0xFF,                             //lui     $t8, XXXX
    0x24, 0x19, 0xFF, 0xFF,                             //li      $t9, XXXX                 ; <---fw_appl_version
    0x25, 0xef, 0xFF, 0xFF,                             //la      $t7, TAP_GetState
    0x25, 0xce, 0xFF, 0xFF,                             //la      $t6, TAP_ExitNormal
    0x27, 0x18, 0xFF, 0xFF,                             //la      $t8, TAP_Exit
    0xac, 0x99, 0x00, 0x2c,                             //sw      $t9, 0x2c($a0)
    0xac, 0x8f, 0x00, 0x34,                             //sw      $t7, 0x34($a0)
    0xac, 0x8e, 0x00, 0x38,                             //sw      $t6, 0x38($a0)
    0xac, 0x98, 0x00, 0x48,                             //sw      $t8, 0x48($a0)
    0x3c, 0x19, 0xFF, 0xFF,                             //lui     $t9, XXXX
    0x3c, 0x0f, 0xFF, 0xFF,                             //lui     $t7, XXXX
    0x3c, 0x0e, 0xFF, 0xFF,                             //lui     $t6, XXXX
};

//for early aus-416-thanks to tonymy01
byte pattern_fw_appl_version3[] = {
    0x24, 0x19, 0xFF, 0xFF,                             //li      $t9, 0x1171               ; <---fw_appl_version
    0x27, 0x18, 0xFF, 0xFF,                             //la      $t8, TAP_SystemProc
    0xac, 0x99, 0x00, 0x2c,                             //sw      $t9, 0x2C($a0)
    0xac, 0x98, 0x00, 0x28,                             //sw      $t8, 0x28($a0)
    0x3c, 0x0e, 0xFF, 0xFF,                             //lui     $t6, XXXX
    0x3c, 0x19, 0xFF, 0xFF,                             //lui     $t9, XXXX
    0x3c, 0x18, 0xFF, 0xFF,                             //lui     $t8, XXXX
    0x25, 0xce, 0xFF, 0xFF,                             //la      $t6, TAP_ExitNormal
    0x27, 0x39, 0xFF, 0xFF,                             //la      $t9, TAP_EnterNormal
    0x27, 0x18, 0xFF, 0xFF,                             //la      $t8, TAP_Exit
    0xac, 0x8e, 0x00, 0x38,                             //sw      $t6, 0x38($a0)
    0xac, 0x99, 0x00, 0x3c,                             //sw      $t9, 0x3C($a0)
    0xac, 0x98, 0x00, 0x48,                             //sw      $t8, 0x48($a0)
    0x3c, 0x0f, 0xFF, 0xFF,                             //lui     $t7, XXXX
    0x3c, 0x0e, 0xFF, 0xFF,                             //lui     $t6, XXXX
    0x3c, 0x19, 0xFF, 0xFF,                             //lui     $t9, XXXX
};

//fw_gp - 0x80000000
byte pattern_fw_gp[] = {
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x3c, 0x08, 0x10, 0x00,                             //lui     $t0, 0x1000
    0x40, 0x88, 0x60, 0x00,                             //mtc0    $t0, SR
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x40, 0x80, 0x68, 0x00,                             //mtc0    $0, Cause
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x3c, 0x1d, 0xFF, 0xFF,                             //li      $sp, XXXXXXXX
    0x27, 0xbd, 0xFF, 0xFF,                             //
    0x3c, 0x1c, 0xFF, 0xFF,                             //li      $gp, XXXXXXXX             ; <---fw_gp
    0x27, 0x9c, 0xFF, 0xFF,                             //
    0x3c, 0x17, 0xFF, 0xFF,                             //la      $s7, XXXXXXXX
    0x26, 0xf7, 0xFF, 0xFF,                             //
    0x3c, 0x08, 0xa0, 0x00,                             //lui     $t0, 0xa000
    0x3c, 0x1b, 0xFF, 0xFF,                             //la      $k1, XXXXXXXX
    0x27, 0x7b, 0xFF, 0xFF,
};

//os_free
byte pattern_fw_os_free[] = {
    0x27, 0xbd, 0xff, 0xc0,                             //addiu   $sp, -0x40
    0xaf, 0xb0, 0x00, 0x20,                             //sw      $s0, 0x20($sp)
    0xaf, 0xb1, 0x00, 0x24,                             //sw      $s1, 0x24($sp)
    0xaf, 0xb2, 0x00, 0x28,                             //sw      $s2, 0x28($sp)
    0xaf, 0xb3, 0x00, 0x2c,                             //sw      $s3, 0x2c($sp)
    0xaf, 0xb4, 0x00, 0x30,                             //sw      $s4, 0x30($sp)
    0xaf, 0xb5, 0x00, 0x34,                             //sw      $s5, 0x34($sp)
    0xaf, 0xb6, 0x00, 0x38,                             //sw      $s6, 0x38($sp)
    0xaf, 0xbe, 0x00, 0x3c,                             //sw      $fp, 0x3c($sp)
    0xaf, 0xbf, 0x00, 0x1c,                             //sw      $ra, 0x1c($sp)
    0x00, 0x80, 0xf0, 0x25,                             //move    $fp, $a0
    0x3c, 0x19, 0xdf, 0xff,                             //li      $t9, 0xdfffffff
    0x37, 0x39, 0xff, 0xff,                             //
    0x03, 0xd9, 0xf0, 0x24,                             //and     $fp, $t9
    0x3c, 0x18, 0xFF, 0xFF,                             //li      $t8, XXXXXXXX
    0x27, 0x18, 0xFF, 0xFF,                             //
};

//fw_add_task_queue
byte pattern_fw_add_task_queue[] = {
    0x27, 0xbd, 0xff, 0xd8,                             //addiu   $sp, -0x28
    0xaf, 0xb3, 0x00, 0x14,                             //sw      $s3, 0x14($sp)
    0xaf, 0xb4, 0x00, 0x18,                             //sw      $s4, 0x18($sp)
    0xaf, 0xb5, 0x00, 0x1c,                             //sw      $s5, 0x1c($sp)
    0xaf, 0xb6, 0x00, 0x20,                             //sw      $s6, 0x20($sp)
    0xaf, 0xbe, 0x00, 0x24,                             //sw      $fp, 0x24($sp)
    0xaf, 0xbf, 0x00, 0x10,                             //sw      $ra, 0x10($sp)
    0x00, 0x80, 0xf0, 0x25,                             //move    $fp, $a0
    0x00, 0xa0, 0xb0, 0x25,                             //move    $s6, $a1
    0x00, 0x00, 0xa0, 0x25,                             //move    $s4, $0
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x40, 0xa8, 0x25,                             //move    $s5, $v0
    0x3c, 0x19, 0xFF, 0xFF,                             //lw      $t9, XXXXXXXX
    0x8f, 0x39, 0xFF, 0xFF,                             //
    0x27, 0x39, 0x00, 0x01,                             //addiu   $t9, 1
};

//fw_event_alloc
byte pattern_fw_event_alloc[] = {
    0x27, 0xbd, 0xff, 0xd0,                             //addiu   $sp, -0x30
    0xaf, 0xbf, 0x00, 0x1c,                             //sw      $ra, 0x1c($sp)
    0xaf, 0xb4, 0x00, 0x20,                             //sw      $s4, 0x20($sp)
    0xaf, 0xb5, 0x00, 0x24,                             //sw      $s5, 0x24($sp)
    0xaf, 0xb6, 0x00, 0x28,                             //sw      $s6, 0x28($sp)
    0xaf, 0xbe, 0x00, 0x2c,                             //sw      $fp, 0x2c($sp)
    0x00, 0x80, 0xf0, 0x25,                             //move    $fp, $a0
    0x00, 0xa0, 0xFF, 0x25,                             //move    $s5, $a1                  ; length
    0x27, 0xa4, 0x00, 0x10,                             //addiu   $a0, $sp, 0x10
    0x27, 0xa5, 0x00, 0x14,                             //addiu   $a1, $sp, 0x14
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX
    0x27, 0xa6, 0x00, 0x18,                             //addiu   $a2, $sp, 0x18
    0x3c, 0xFF, 0xFF, 0xFF,                             //la      $XX, event_table
    0x26, 0xFF, 0xFF, 0xFF,                             //
    0x00, 0x1e, 0xFF, 0x00,                             //sll     $XX, $fp, 4
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     os_malloc
};

//fw_event_delete
byte pattern_fw_event_delete[] = {
    0x3c, 0x06, 0xFF, 0xFF,                             //lui     $a2, XXXX
    0x27, 0xbd, 0xff, 0xe0,                             //addiu   $sp, -0x20
    0x24, 0xc6, 0xFF, 0xFF,                             //la      $a2, event_table
    0x00, 0x04, 0xc1, 0x80,                             //sll     $t8, $a0, 6
    0xaf, 0xbe, 0x00, 0x1c,                             //sw      $fp, 0x1c($sp)
    0xaf, 0xbf, 0x00, 0x18,                             //sw      $ra, 0x18($sp)
    0x00, 0xd8, 0xc0, 0x21,                             //addu    $t8, $a2, $t8
    0x8f, 0x18, 0x00, 0x00,                             //lw      $t8, 0($t8)
    0x00, 0x18, 0xc7, 0x82,                             //srl     $t8, 30
    0x13, 0x00, 0x00, 0x1a,                             //beqz    $t8, XXXX
    0x00, 0x80, 0xf0, 0x25,                             //move    $fp, $a0
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x1e, 0xc9, 0x80,                             //sll     $t9, $fp, 6
    0x00, 0xd9, 0xf0, 0x21,                             //addu    $fp, $a2, $t9
    0x93, 0xd8, 0x00, 0x34,                             //lbu     $t8, 0x34($fp)
};

//1288
//fw_event_delete
byte pattern_fw_event_delete_1288[] = {
    0x27, 0xbd, 0xff, 0xe0,                             //addiu   $sp, -0x20
    0x24, 0x18, 0x00, 0x44,                             //li      $t8, 0x44
    0xaf, 0xbe, 0x00, 0x1c,                             //sw      $fp, 0x1c($sp)
    0x3c, 0x06, 0xFF, 0xFF,                             //lui     $a2, XXXX
    0x00, 0x98, 0x00, 0x19,                             //multu   $a0, $t8
    0x24, 0xc6, 0xFF, 0xFF,                             //la      $a2, XXXX
    0x00, 0x00, 0xf0, 0x12,                             //mflo    $fp
    0xaf, 0xbf, 0x00, 0x18,                             //sw      $ra, 0x18($sp)
    0x00, 0xde, 0x78, 0x21,                             //addu    $t7, $a2, $fp
    0x8d, 0xef, 0x00, 0x00,                             //lw      $t7, 0($t7)
    0x00, 0x0f, 0x7f, 0x82,                             //srl     $t7, 30
    0x51, 0xe0, 0x00, 0x1a,                             //beqzl   $t7, XXXX
    0x8f, 0xbf, 0x00, 0x18,                             //lw      $ra, 0x18($sp)
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0xde, 0xf0, 0x21,                             //addu    $fp, $a2, $fp
    0x93, 0xd8, 0x00, 0x34,                             //lbu     $t8, 0x34($fp)
};

//fw_utctolocal_time
byte pattern_fw_utctolocal_time[] = {
    0x27, 0xbd, 0xff, 0xd0,                             //addiu   $sp, -0x30
    0xaf, 0xbf, 0x00, 0x28,                             //sw      $ra, 0x28($sp)
    0xaf, 0xbe, 0x00, 0x2c,                             //sw      $fp, 0x2c($sp)
    0x00, 0x80, 0xf0, 0x25,                             //move    $fp, $a0
    0x27, 0xa4, 0x00, 0x24,                             //addiu   $a0, $sp, 0x24
    0x27, 0xa5, 0x00, 0x20,                             //addiu   $a1, $sp, 0x20
    0x27, 0xa6, 0x00, 0x21,                             //addiu   $a2, $sp, 0x21
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     get_time
    0x27, 0xa7, 0x00, 0x22,                             //addiu   $a3, $sp, 0x22
    0x97, 0xb9, 0x00, 0x24,                             //lhu     $t9, 0x24($sp)
    0x93, 0xb8, 0x00, 0x20,                             //lbu     $t8, 0x20($sp)
    0x8f, 0x83, 0xFF, 0xFF,                             //lw      $v1, XXXX($gp)
    0x00, 0x19, 0xcc, 0x00,                             //sll     $t9, 16
    0x00, 0x18, 0xc2, 0x00,                             //sll     $t8, 8
    0x93, 0xaf, 0x00, 0x21,                             //lbu     $t7, 0x21($sp)
    0x90, 0x6e, 0x00, 0x04,                             //lbu     $t6, 4($v1)
};

//fw_process_dsmcc
byte pattern_fw_process_dsmcc[] = {
    0x27, 0xbd, 0xff, 0x78,                             //addiu   $sp, -0x88
    0xaf, 0xb6, 0x00, 0x80,                             //sw      $s6, 0x80($sp)
    0xaf, 0xbe, 0x00, 0x84,                             //sw      $fp, 0x84($sp)
    0xaf, 0xbf, 0x00, 0x7c,                             //sw      $ra, 0x7c($sp)
    0x00, 0x80, 0x30, 0x25,                             //move    $a2, $a0
    0x90, 0x9e, 0x00, 0x00,                             //lbu     $fp, 0($a0)
    0x00, 0xa0, 0xb0, 0x25,                             //move    $s6, $a1
    0x27, 0xa4, 0x00, 0x24,                             //addiu   $a0, $sp, 0x24
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX
    0x00, 0xc0, 0x28, 0x25,                             //move    $a1, $a2
    0x50, 0x40, 0x00, 0x3b,                             //beqzl   $v0, XXXX
    0x8f, 0xbf, 0x00, 0x7c,                             //lw      $ra, 0x7c($sp)
    0x34, 0x01, 0x00, 0x3b,                             //li      $1, 0x3b                  ; user_network_message_section
    0x13, 0xc1, 0x00, 0x09,                             //beq     $fp, $1, XXXX
    0x27, 0xa4, 0x00, 0x28,                             //addiu   $a0, $sp, 0x28
    0x34, 0x01, 0x00, 0x3c,                             //li      $1, 0x3c                  ; download_data_message_section
};

//fw_process_dsmcc
byte pattern_fw_process_dsmcc2[] = {
    0x27, 0xbd, 0xff, 0xc8,                             //addiu   $sp, -0x38
    0xaf, 0xb3, 0x00, 0x24,                             //sw      $s3, 0x24($sp)
    0xaf, 0xb4, 0x00, 0x28,                             //sw      $s4, 0x28($sp)
    0xaf, 0xb5, 0x00, 0x2c,                             //sw      $s5, 0x2c($sp)
    0xaf, 0xb6, 0x00, 0x30,                             //sw      $s6, 0x30($sp)
    0xaf, 0xbe, 0x00, 0x34,                             //sw      $fp, 0x34($sp)
    0xaf, 0xbf, 0x00, 0x20,                             //sw      $ra, 0x20($sp)
    0x90, 0x99, 0x00, 0x01,                             //lbu     $t9, 1($a0)
    0x90, 0x98, 0x00, 0x02,                             //lbu     $t8, 2($a0)
    0x33, 0x39, 0x00, 0x0f,                             //andi    $t9, 0xf
    0x00, 0x19, 0xca, 0x00,                             //sll     $t9, 8
    0x03, 0x38, 0xf0, 0x25,                             //or      $fp, $t9, $t8
    0x27, 0xde, 0x00, 0x03,                             //addiu   $fp, 3
    0x33, 0xd3, 0xff, 0xff,                             //andi    $s3, $fp, 0xffff
    0x00, 0x80, 0xb0, 0x25,                             //move    $s6, $a0
    0x00, 0xa0, 0xa8, 0x25,                             //move    $s5, $a1
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX
    0x02, 0x60, 0x28, 0x25,                             //move    $a1, $s3
    0x3c, 0x1e, 0x40, 0x05,                             //li      $fp, 0x40051000
    0x37, 0xde, 0x10, 0x00,                             //
    0x3c, 0x05, 0xFF, 0xFF,                             //lui     $a1, XXXX
    0x00, 0x40, 0xa0, 0x25,                             //move    $s4, $v0
    0x24, 0xa5, 0xFF, 0xFF,                             //la      $a1, XXXX
};

//fw_process_dsmcc
byte pattern_fw_process_dsmcc3[] = {
    0x27, 0xbd, 0xff, 0xe0,                             //addiu   $sp, -0x20
    0xaf, 0xb6, 0x00, 0x18,                             //sw      $s6, 0x18($sp)
    0xaf, 0xbe, 0x00, 0x1c,                             //sw      $fp, 0x1c($sp)
    0xaf, 0xbf, 0x00, 0x14,                             //sw      $ra, 0x14($sp)
    0x90, 0x99, 0x00, 0x01,                             //lbu     $t9, 1($a0)
    0x90, 0x98, 0x00, 0x02,                             //lbu     $t8, 2($a0)
    0x33, 0x39, 0x00, 0x0f,                             //andi    $t9, 0xf
    0x00, 0x19, 0xca, 0x00,                             //sll     $t9, 8
    0x03, 0x38, 0x10, 0x25,                             //or      $v0, $t9, $t8
    0x24, 0x42, 0x00, 0x03,                             //addiu   $v0, 3
    0x30, 0x42, 0xff, 0xff,                             //andi    $v0, 0xffff               ; section_len
    0x00, 0x80, 0xf0, 0x25,                             //move    $fp, $a0                  ; section_ptr
    0x00, 0xa0, 0xb0, 0x25,                             //move    $s6, $a1
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     dsmcc_crc32
    0x00, 0x40, 0x28, 0x25,                             //move    $a1, $v0
    0x54, 0x40, 0x00, 0x18,                             //bnezl   $v0, XXXX
    0x8f, 0xbf, 0x00, 0x14,                             //lw      $ra, 0x14($sp)
    0x93, 0xcd, 0x00, 0x00,                             //lbu     $t5, 0x00($fp)            ; table_id
    0x34, 0x01, 0x00, 0x3b,                             //li      $1, 0x3b                  ; user_network_message_section
    0x11, 0xa1, 0x00, 0x09,                             //beq     $t5, $1, XXXX
    0x03, 0xc0, 0x20, 0x25,                             //move    $a0, $fp
    0x34, 0x01, 0x00, 0x3c,                             //li      $1, 0x3c                  ; download_data_message_section
};

//fw_process_eit
byte pattern_fw_process_eit[] = {
    0x27, 0xbd, 0xff, 0xFF,                             //addiu   $sp, -XX
    0xaf, 0xb0, 0x00, 0xFF,                             //sw      $s0, XX($sp)
    0xaf, 0xb1, 0x00, 0xFF,                             //sw      $s1, XX($sp)
    0xaf, 0xb2, 0x00, 0xFF,                             //sw      $s2, XX($sp)

    0xaf, 0xb3, 0x00, 0xFF,                             //sw      $s3, XX($sp)
    0xaf, 0xb4, 0x00, 0xFF,                             //sw      $s4, XX($sp)
    0xaf, 0xb5, 0x00, 0xFF,                             //sw      $s5, XX($sp)
    0xaf, 0xb6, 0x00, 0xFF,                             //sw      $s6, XX($sp)

    0xaf, 0xbe, 0x00, 0xFF,                             //sw      $fp, XX($sp)
    0xaf, 0xbf, 0x00, 0xFF,                             //sw      $ra, XX($sp)
    0xaf, 0xa4, 0x00, 0xFF,                             //sw      $a0, XX($sp)
    0x24, 0x19, 0x00, 0xff,                             //li      $t9, 0xff

    0xa3, 0xb9, 0x00, 0xFF,                             //sb      $t9, XX($sp)
    0xa3, 0xa0, 0x00, 0xFF,                             //sb      $0, XX($sp)
};

//fw_shutdown_task
byte pattern_fw_shutdown_task[] = {
    0x27, 0xbd, 0xff, 0xb8,                             //addiu   $sp, -0x48
    0xaf, 0xb4, 0x00, 0x38,                             //sw      $s4, 0x38($sp)
    0xaf, 0xb5, 0x00, 0x3c,                             //sw      $s5, 0x3c($sp)
    0xaf, 0xb6, 0x00, 0x40,                             //sw      $s6, 0x40($sp)
    0xaf, 0xbe, 0x00, 0x44,                             //sw      $fp, 0x44($sp)
    0xaf, 0xbf, 0x00, 0x34,                             //sw      $ra, 0x34($sp)
    0x24, 0x19, 0x00, 0xff,                             //li      $t9, 0xff
    0xaf, 0xa5, 0x00, 0x4c,                             //sw      $a1, 0x4c($sp)
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX
    0xa3, 0xb9, 0x00, 0x2c,                             //sb      $t9, 0x2c($sp)
    0x83, 0x98, 0xFF, 0xFF,                             //lb      $t8, XXXX($gp)
    0x24, 0x0f, 0x00, 0x01,                             //li      $t7, 1
    0x57, 0x0f, 0x00, 0x03,                             //bnel    $t8, $t7, XXXX
    0x93, 0x96, 0xFF, 0xFF,                             //lbu     $s6, XXXX($gp)
    0x10, 0x00, 0x00, 0xff,                             //b       XXXX
    0x00, 0x00, 0x10, 0x25,                             //move    $v0, $0
};

//fw_shutdown_irq_disable
byte pattern_fw_shutdown_irq_disable[] = {
    0x27, 0xbd, 0xff, 0xd8,                             //addiu   $sp, -0x28
    0xaf, 0xbf, 0x00, 0x24,                             //sw      $ra, 0x24($sp)
    0x24, 0x19, 0x00, 0x0a,                             //li      $t9, 0xa
    0xaf, 0xb9, 0x00, 0x10,                             //sw      $t9, 0x10($sp)
    0x27, 0xa5, 0x00, 0x10,                             //addiu   $a1, $sp, 0x10
    0x00, 0x00, 0x20, 0x25,                             //move    $a0, $0
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     ioctl
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x00, 0x10, 0x25,                             //move    $v0, $0
    0x8f, 0xbf, 0x00, 0x24,                             //lw      $ra, 0x24($sp)
    0x27, 0xbd, 0x00, 0x28,                             //addiu   $sp, 0x28
    0x03, 0xe0, 0x00, 0x08,                             //jr      $ra
    0x00, 0x00, 0x00, 0x00,                             //nop
};

//fw_shutdown_irq_enable
byte pattern_fw_shutdown_irq_enable[] = {
    0x27, 0xbd, 0xff, 0xd8,                             //addiu   $sp, -0x28
    0xaf, 0xbf, 0x00, 0x24,                             //sw      $ra, 0x24($sp)
    0x24, 0x19, 0x00, 0x0b,                             //li      $t9, 0xb
    0xaf, 0xb9, 0x00, 0x10,                             //sw      $t9, 0x10($sp)
    0x27, 0xa5, 0x00, 0x10,                             //addiu   $a1, $sp, 0x10
    0x00, 0x00, 0x20, 0x25,                             //move    $a0, $0
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     ioctl
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x00, 0x10, 0x25,                             //move    $v0, $0
    0x8f, 0xbf, 0x00, 0x24,                             //lw      $ra, 0x24($sp)
    0x27, 0xbd, 0x00, 0x28,                             //addiu   $sp, 0x28
    0x03, 0xe0, 0x00, 0x08,                             //jr      $ra
    0x00, 0x00, 0x00, 0x00,                             //nop
};

//TAP_KbHit
byte pattern_TAP_KbHit[] = {
    0x27, 0xbd, 0xff, 0xe0,                             //addiu   $sp, -0x20                ;  many tap functions
    0xaf, 0xbe, 0x00, 0x1c,                             //sw      $fp, 0x1c($sp)
    0x3c, 0x1e, 0xFF, 0xFF,                             //lui     $fp, XXXX
    0xaf, 0xbf, 0x00, 0x14,                             //sw      $ra, 0x14($sp)
    0x27, 0xde, 0xFF, 0xFF,                             //la      $fp, TapIdent
    0xaf, 0xb6, 0x00, 0x18,                             //sw      $s6, 0x18($sp)
    0x8f, 0xcf, 0x00, 0x00,                             //lw      $t7, 0x00($fp)
    0x3c, 0x16, 0xFF, 0xFF,                             //la      $s6, TapTable
    0x26, 0xd6, 0xFF, 0xFF,                             //
    0x00, 0x0f, 0x79, 0x40,                             //sll     $t7, 5
    0x02, 0xcf, 0x78, 0x21,                             //addu    $t7, $s6, $t7
    0x8d, 0xe3, 0x00, 0x00,                             //lw      $v1, 0($t7)
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x03, 0x80, 0x10, 0x25,                             //move    $v0, $gp
    0x00, 0x00, 0x00, 0x00,                             //nop
};

//TAP_GetCurrentEvent
byte pattern_TAP_GetCurrentEvent[] = {
    0x27, 0xbd, 0xff, 0xb8,                             //addiu   $sp, -0x48
    0xaf, 0xb1, 0x00, 0x2c,                             //sw      $s1, 0x2c($sp)
    0xaf, 0xb2, 0x00, 0x30,                             //sw      $s2, 0x30($sp)
    0xaf, 0xb3, 0x00, 0x34,                             //sw      $s3, 0x34($sp)
    0xaf, 0xb4, 0x00, 0x38,                             //sw      $s4, 0x38($sp)
    0xaf, 0xb5, 0x00, 0x3c,                             //sw      $s5, 0x3c($sp)
    0xaf, 0xb6, 0x00, 0x40,                             //sw      $s6, 0x40($sp)
    0xaf, 0xbe, 0x00, 0x44,                             //sw      $fp, 0x44($sp)
    0x3c, 0x11, 0xFF, 0xFF,                             //lui     $s1, XXXX
    0xaf, 0xbf, 0x00, 0x28,                             //sw      $ra, 0x48+var_20($sp)
    0x26, 0x31, 0xFF, 0xFF,                             //la      $s1, TapIdent
    0x8e, 0xFF, 0x00, 0x00,                             //lw      $XX, 0($s1)
    0x3c, 0xFF, 0xFF, 0xFF,                             //la      $XX, TapTable
    0xff, 0xFF, 0xFF, 0xFF,                             //
    0x00, 0xFF, 0xFF, 0x40,                             //sll     $XX, 5
    0xff, 0xFF, 0xFF, 0x21,                             //addu    $XX, $XX

    0xFF, 0xFF, 0x00, 0x00,                             //lw      $XX, 0($XX)
    0x00, 0x80, 0xa0, 0x25,                             //move    $s4, $a0
    0x00, 0xa0, 0xa8, 0x25,                             //move    $s5, $a1
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x03, 0x80, 0x10, 0x25,                             //move    $v0, $gp
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x02, 0xc0, 0xe0, 0x25,                             //move    $gp, $s6
    0x3c, 0x16, 0xFF, 0xFF,                             //li      $s6, XXXXXXXX
    0x26, 0xd6, 0xFF, 0xFF,                             //
    0x02, 0xc0, 0x20, 0x25,                             //move    $a0, $s6
    0x00, 0x00, 0x28, 0x25,                             //move    $a1, $0
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX                    ; setmem
    0x24, 0x06, 0xFF, 0xFF,                             //li      $a2, XXXX                 ; sizeof(type_tapevent), 0x118
    0x3c, 0x1e, 0xFF, 0xFF,                             //la      $fp, XXXXXXXX             ; event_table

    0x27, 0xde, 0xFF, 0xFF,                             //
    0x24, 0x13, 0xFF, 0xFF,                             //li      $s3, XXXX                 ; 1388 = 5000 = MAX_EVENTS
    0x8f, 0xc5, 0x00, 0x00,                             //lw      $a1, 0x00($fp)
    0x24, 0x0f, 0x00, 0x04,                             //li      $t7, 4

    0x00, 0x05, 0xc0, 0xFF,                             //sll     $t8, $a1, 3               ;2 from 1288
    0x00, 0x18, 0xFF, 0xFF,                             //srl     $t8, 27                   ;29 from 1288
    0x57, 0x0f, 0xFF, 0xFF,                             //bnel    $t8, $t7, XXXX
    0x27, 0xde, 0x00, 0xFF,                             //addiu   $fp, 0x40                 ;44 from 1288
};

//process_tdt
byte pattern_process_tdt[] = {
    0x27, 0xbd, 0xff, 0xc0,                             //addiu   $sp, -0x40
    0xaf, 0xbf, 0x00, 0x30,                             //sw      $ra, 0x30($sp)
    0xaf, 0xb5, 0x00, 0x34,                             //sw      $s5, 0x34($sp)
    0xaf, 0xb6, 0x00, 0x38,                             //sw      $s6, 0x38($sp)
    0xaf, 0xbe, 0x00, 0x3c,                             //sw      $fp, 0x3c($sp)
    0x27, 0xa5, 0x00, 0x21,                             //addiu   $a1, $sp, 0x21
    0x27, 0xa6, 0x00, 0x20,                             //addiu   $a2, $sp, 0x20
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     dbgsi_loaddmx
    0x27, 0xa7, 0x00, 0x24,                             //addiu   $a3, $sp, 0x24
    0x93, 0xae, 0x00, 0x20,                             //lbu     $t6, 0x20($sp)
    0x24, 0x19, 0x00, 0x01,                             //li      $t9, 1
    0x55, 0xd9, 0xFF, 0xFF,                             //bnel    $t6, $t9, XXXX
    0x8f, 0xbf, 0x00, 0x30,                             //lw      $ra, 0x30($sp)
    0x30, 0xFF, 0xff, 0xff,                             //andi    $XX, $v0, 0xffff
    0xFF, 0xFF, 0xFF, 0xFF,                             //bnezl   $XX, XXXX
    0x8f, 0xbf, 0x00, 0x30,                             //lw      $ra, 0x30($sp)
    0x8f, 0xb9, 0x00, 0x24,                             //lw      $t9, 0x24($sp)
    0x8f, 0xFF, 0x00, 0x00,                             //lw      $s5, 0($t9)
    0xFF, 0xFF, 0xFF, 0xFF,                             //beqzl   $s5, XXXX
    0x8f, 0xbf, 0x00, 0x30,                             //lw      $ra, 0x30($sp)
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     load12                    ; section_length
    0xFF, 0xFF, 0x00, 0x01,                             //addiu   $a0, $s5, 1
    0x24, 0x18, 0x00, 0x05,                             //li      $t8, 5
    0x54, 0x58, 0xFF, 0xFF,                             //bnel    $v0, $t8, XXXX
    0x8f, 0xbf, 0x00, 0x30,                             //lw      $ra, 0x30($sp)
    0xFF, 0xFF, 0xFF, 0xFF,                             //lw      $ra, 0x30($sp)
    0xFF, 0xFF, 0xFF, 0xFF,                             //lbu     $fp, XXXX($gp)
    0xFF, 0xFF, 0xFF, 0xFF,                             //li      $s6, XXXXXXXX             ;  <--utc time
};
//(utc time is different for early aus-416)

//8010DF78            - e.g. GMT: 2006/9/8(Fri) 20:14
byte pattern_fw_pr_gmt[] = {
    0x27, 0xbd, 0xff, 0xe8,                             //addiu   $sp, -0x18
    0xaf, 0xbe, 0x00, 0x14,                             //sw      $fp, 0x18+var_4($sp)
    0xaf, 0xbf, 0x00, 0x10,                             //sw      $ra, 0x18+var_8($sp)
    0x3c, 0x04, 0xFF, 0xFF,                             //lui     $a0, XXXX
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     print_uart_s
    0x24, 0x84, 0xFF, 0xFF,                             //la      $a0, "GMT: "
    0x8f, 0x98, 0xFF, 0xFF,                             //lw      $t8, -0x7B40($gp)         ;  <--utc_time
    0x00, 0x18, 0xc4, 0x02,                             //srl     $t8, 16
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX
    0x33, 0x04, 0xff, 0xff,                             //andi    $a0, $t8, 0xFFFF
    0x8f, 0x9e, 0xFF, 0xFF,                             //lw      $fp, -0x7B40($gp)
    0x3c, 0x04, 0xFF, 0xFF,                             //lui     $a0, XXXX
    0x00, 0x1e, 0x7a, 0x02,                             //srl     $t7, $fp, 8
    0x31, 0xe5, 0x00, 0xff,                             //andi    $a1, $t7, 0xFF
    0x24, 0x84, 0xFF, 0xFF,                             //la      $a0, " %02d:%02d\n"
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     rs232_printf
};

//process_tot
byte pattern_process_tot[] = {
    0x27, 0xbd, 0xff, 0xc8,                             //addiu   $sp, -0x38
    0xaf, 0xbf, 0x00, 0x2c,                             //sw      $ra, 0x38+var_c($sp)
    0xaf, 0xb6, 0x00, 0x30,                             //sw      $s6, 0x38+var_8($sp)
    0xaf, 0xbe, 0x00, 0x34,                             //sw      $fp, 0x38+var_4($sp)
    0x27, 0xa5, 0x00, 0x21,                             //addiu   $a1, $sp, 0x38+var_18+1
    0x27, 0xa6, 0x00, 0x20,                             //addiu   $a2, $sp, 0x38+var_18
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     dbgsi_loaddmx
    0x27, 0xa7, 0x00, 0x24,                             //addiu   $a3, $sp, 0x38+var_14
    0x93, 0xae, 0x00, 0x20,                             //lbu     $t6, 0x38+var_18($sp)
    0x24, 0x19, 0x00, 0x01,                             //li      $t9, 1
    0x55, 0xd9, 0x00, 0x29,                             //bnel    $t6, $t9, loc_8010fea0
    0x8f, 0xbf, 0x00, 0x2c,                             //lw      $ra, 0x38+var_c($sp)
    0x30, 0x56, 0xff, 0xff,                             //andi    $s6, $v0, 0xffff
    0x56, 0xc0, 0x00, 0x26,                             //bnezl   $s6, loc_8010fea0
    0x8f, 0xbf, 0x00, 0x2c,                             //lw      $ra, 0x38+var_c($sp)
    0x8f, 0xb9, 0x00, 0x24,                             //lw      $t9, 0x38+var_14($sp)
//    0x8f, 0x36, 0x00, 0x00,                             //lw      $s6, 0($t9)
//    0x52, 0xc0, 0x00, 0x22,                             //beqzl   $s6, loc_8010fea0
//    0x8f, 0xbf, 0x00, 0x2c,                             //lw      $ra, 0x38+var_c($sp)
//    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     load12           # section_length
//    0x26, 0xc4, 0x00, 0x01,                             //addiu   $a0, $s6, 1
//    0x24, 0x5e, 0x00, 0x03,                             //addiu   $fp, $v0, 3
//    0x00, 0x1e, 0xf4, 0x00,                             //sll     $fp, 16
//    0x00, 0x1e, 0xf4, 0x03,                             //sra     $fp, 16
//    0x2b, 0xc1, 0x00, 0x08,                             //slti    $1, $fp, 8
//    0x14, 0x20, 0x00, 0x19,                             //bnez    $1, loc_8010fe9c
//    0x00, 0x00, 0x00, 0x00,                             //nop
//    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     load12           # descriptors_loop_length
};

byte pattern_process_tot2[] = {
    0x27, 0xbd, 0xff, 0xc0,                             //addiu   $sp, -0x40
    0xaf, 0xbf, 0x00, 0x28,                             //sw      $ra, 0x40+var_18($sp)
    0xaf, 0xb3, 0x00, 0x2c,                             //sw      $s3, 0x40+var_14($sp)
    0xaf, 0xb4, 0x00, 0x30,                             //sw      $s4, 0x40+var_10($sp)
    0xaf, 0xb5, 0x00, 0x34,                             //sw      $s5, 0x40+var_C($sp)
    0xaf, 0xb6, 0x00, 0x38,                             //sw      $s6, 0x40+var_8($sp)
    0xaf, 0xbe, 0x00, 0x3c,                             //sw      $fp, 0x40+var_4($sp)
    0x27, 0xa5, 0x00, 0x21,                             //addiu   $a1, $sp, 0x40+var_1F
    0x27, 0xa6, 0x00, 0x20,                             //addiu   $a2, $sp, 0x20
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     dbgsi_loaddmx
    0x27, 0xa7, 0x00, 0x24,                             //addiu   $a3, $sp, 0x40+var_1C
    0x93, 0xb9, 0x00, 0x20,                             //lbu     $t9, 0x40+var_20($sp)
    0x24, 0x18, 0x00, 0x01,                             //li      $t8, 1
    0x57, 0x38, 0x00, 0x59,                             //bnel    $t9, $t8, loc_8011324C
    0x8f, 0xbf, 0x00, 0x28,                             //lw      $ra, 0x40+var_18($sp)
    0x30, 0x59, 0xff, 0xff,                             //andi    $t9, $v0, 0xFFFF

};

//fw_display_func
byte pattern_fw_display_func[] = {
    0x27, 0xbd, 0xff, 0xd8,                             //addiu   $sp, -0x28
    0xaf, 0xb6, 0x00, 0x20,                             //sw      $s6, 0x20($sp)
    0xaf, 0xbe, 0x00, 0x24,                             //sw      $fp, 0x24($sp)
    0xaf, 0xbf, 0x00, 0x1c,                             //sw      $ra, 0x1c($sp)
    0x00, 0x80, 0xf0, 0x25,                             //move    $fp, $a0
    0x00, 0xa0, 0xb0, 0x25,                             //move    $s6, $a1
    0x33, 0xd9, 0x00, 0x01,                             //andi    $t9, $fp, 1
    0x13, 0x20, 0x00, 0x03,                             //beqz    $t9, XXXX
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x92, 0xd9, 0x00, 0x00,                             //lbu     $t9, 0($s6)
    0xa3, 0x99, 0xFF, 0xFF,                             //sb      $t9, XXXX($gp)
    0x33, 0xd9, 0x00, 0x02,                             //andi    $t9, $fp, 2
    0x13, 0x20, 0x00, 0x03,                             //beqz    $t9, XXXX
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x92, 0xd9, 0x00, 0x01,                             //lbu     $t9, 1($s6)
    0xa3, 0x99, 0xFF, 0xFF,                             //sb      $t9, XXXX($gp)
};

//fw_translate_and_display
byte pattern_fw_translate_and_display[] = {
    0xa0, 0xc0, 0x00, 0x00,                             //sb      $0, 0($a2)
    0x24, 0xa5, 0xFF, 0xFF,                             //addiu   $a1, 0xFFFF
    0x10, 0xa0, 0x00, 0x07,                             //beqz    $a1, XXXX
    0x24, 0xc6, 0x00, 0x01,                             //addiu   $a2, 1
    0x00, 0x05, 0x28, 0x42,                             //srl     $a1, 1
    0xa0, 0xc0, 0x00, 0x00,                             //sb      $0, 0($a2)
    0xa0, 0xc0, 0x00, 0x01,                             //sb      $0, 1($a2)
    0x24, 0xa5, 0xFF, 0xFF,                             //addiu   $a1, 0xFFFF
    0x14, 0xa0, 0xff, 0xfc,                             //bnez    $a1, XXXX
    0x24, 0xc6, 0x00, 0x02,                             //addiu   $a2, 2
    0x27, 0x85, 0xFF, 0xFF,                             //addiu   $a1, $gp, XXXX
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     display_func
    0x24, 0x04, 0x00, 0x0f,                             //li      $a0, 0xf
    0x8f, 0xbf, 0x00, 0x14,                             //lw      $ra, 0x14($sp)
    0x03, 0xe0, 0x00, 0x08,                             //jr      $ra
    0x27, 0xbd, 0x00, 0x18,                             //addiu   $sp, 0x18
};

//fw_hdd_move
byte pattern_fw_hdd_move[] = {
    0x27, 0xbd, 0xff, 0xd8,                             //addiu   $sp, -0x28                ; move(dir_type *from, dir_type *to, char * to_filename);
    0xaf, 0xb4, 0x00, 0x18,                             //sw      $s4, 0x18($sp)
    0xaf, 0xb5, 0x00, 0x1c,                             //sw      $s5, 0x1c($sp)
    0xaf, 0xb6, 0x00, 0x20,                             //sw      $s6, 0x20($sp)
    0xaf, 0xbe, 0x00, 0x24,                             //sw      $fp, 0x24($sp)
    0xaf, 0xbf, 0x00, 0x14,                             //sw      $ra, 0x14($sp)
    0x00, 0x80, 0xf0, 0x25,                             //move    $fp, $a0
    0x00, 0xa0, 0xb0, 0x25,                             //move    $s6, $a1
    0x00, 0xc0, 0xa8, 0x25,                             //move    $s5, $a2
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x30, 0x42, 0xff, 0xff,                             //andi    $v0, 0xffff
    0x10, 0x40, 0x00, 0x04,                             //beqz    $v0, XXXX
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x34, 0x02, 0xfe, 0x00,                             //li      $v0, 0xfe00
    0x10, 0x00, 0x00, 0x19,                             //b       XXXX
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x02, 0xa0, 0x20, 0x25,                             //move    $a0, $s5
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXX                      ; to_legal_filename
    0x00, 0x00, 0x00, 0x00,                             //nop
};

//fw_mheg_entry
byte pattern_fw_mheg_entry[] = {
    0x53, 0xc0, 0x00, 0x7c,                             //beqzl   $fp, XXXX
    0x8f, 0x99, 0xFF, 0xFF,                             //lw      $t9, -0x74D8($gp)
    0x8f, 0x99, 0xFF, 0xFF,                             //lw      $t9, -0x74F8($gp)
    0x57, 0x20, 0x00, 0x79,                             //bnezl   $t9, XXXX
    0x8f, 0x99, 0xFF, 0xFF,                             //lw      $t9, -0x74D8($gp)
    0x8f, 0x83, 0xFF, 0xFF,                             //lw      $v1, -0x74F4($gp)         ; <---
    0x10, 0x60, 0x00, 0x15,                             //beqz    $v1, XXXX
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x8c, 0x79, 0x00, 0x08,                             //lw      $t9, 8($v1)
    0x8f, 0x22, 0x00, 0x34,                             //lw      $v0, 0x34($t9)            ; register_group
    0x24, 0x18, 0x00, 0x03,                             //li      $t8, 3                    ; register_group_3
    0x14, 0x58, 0x00, 0x05,                             //bne     $v0, $t8, XXXX
    0x24, 0x19, 0x00, 0x05,                             //li      $t9, 5                    ; register_group_5
    0x2b, 0xc1, 0x00, 0x10,                             //slti    $1, $fp, 0x10
    0x54, 0x20, 0x00, 0x6e,                             //bnezl   $1, XXXX
    0x8f, 0x99, 0xFF, 0xFF,                             //lw      $t9, -0x74D8($gp)
};

//fw_mheg_entry
byte pattern_fw_mheg_entry2[] = {
    0x27, 0xa6, 0x00, 0x3c,                             //addiu   $a2, $sp, 0x3c
    0x0c, 0xFF, 0xFF, 0xff,                             //jal     XXXXXX
    0x24, 0x05, 0x00, 0x1f,                             //li      $a1, 0x1f
    0x8f, 0x9e, 0xFF, 0xFF,                             //lw      $fp, -0x750C($gp)
    0x53, 0xc0, 0x00, 0x0f,                             //beqzl   $fp, XXXX
    0x8f, 0x9e, 0xFF, 0xFF,                             //lw      $fp, -0x7510($gp)         ; <---
    0x8f, 0xd9, 0x00, 0x08,                             //lw      $t9, 0x0($fp)
    0x8f, 0x39, 0x00, 0x34,                             //lw      $t9, 0x34($t9)            ; register_group
    0x24, 0x18, 0x00, 0x03,                             //li      $t8, 3
    0x57, 0x38, 0x00, 0x06,                             //bnel    $t9, $t8, XXXX
    0x8f, 0xc4, 0x00, 0x00,                             //lw      $a0, 0x00($fp)
    0x8f, 0xb9, 0x00, 0x24,                             //lw      $t9, 0x24($sp)
    0x2b, 0x21, 0x00, 0x64,                             //slti    $1, $t9, 0x64
    0x14, 0x20, 0xFF, 0xFF,                             //bnez    $1, XXXX
    0x27, 0xa6, 0x00, 0x2c,                             //addiu   $a2, $sp, 0x2c
    0x8f, 0xc4, 0x00, 0x00,                             //lw      $a0, 0x00($fp)
};

//fw_key_map_patch                                       0x80131454
byte pattern_fw_key_map_patch[] = {
    0xaf, 0x22, 0x00, 0x00,                             //sw      $v0, 0($t9)
    0xaf, 0x0f, 0xFF, 0xFF,                             //sw      $t7, TapIdent
    0x00, 0x00, 0xa0, 0x25,                             //move    $s4, $0
    0x26, 0xb5, 0xFF, 0xFF,                             //la      $s5, TapTable

    0x8e, 0xb2, 0x00, 0x10,                             //lw      $s2, 0x10($s5)
    0x52, 0x40, 0x00, 0x19,                             //beqzl   $s2, XXXX
    0x26, 0xb5, 0x00, 0x20,                             //addiu   $s5, 0x20
    0x3c, 0x19, 0xFF, 0xFF,                             //lui     $t9, XXXX

    0xaf, 0x34, 0xFF, 0xFF,                             //sw      $s4, TapIdent
    0x8e, 0x50, 0x00, 0x1c,                             //lw      $s0, 0x1c($s2)
    0x8f, 0xd1, 0x00, 0x00,                             //lw      $s1, 0x00($fp)
    0x8e, 0xb2, 0x00, 0x04,                             //lw      $s2, 4($s5)

    0x03, 0xc0, 0x78, 0x25,                             //move    $t7, $fp
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x03, 0x80, 0x10, 0x25,                             //move    $v0, $gp

    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x02, 0x40, 0xe0, 0x25,                             //move    $gp, $s2
    0x24, 0x04, 0x01, 0x00,                             //li      $a0, 0x100                ; evt_key

    0x02, 0x20, 0x28, 0x25,                             //move    $a1, $s1
    0x02, 0x00, 0xf8, 0x09,                             //jalr    $s0
//58
    0x00, 0x00, 0x30, 0x25,                             //move    $a2, $0                   ;<---key_map_patch2: {0x02, 0xc0, 0x30, 0x25  move $a2}, $s6  ; param2 = unmpapped key code
    0xaf, 0xc2, 0x00, 0x00,                             //sw      $v0, 0x00($fp)
    0x8e, 0xb2, 0x00, 0x00,                             //lw      $s2, 0($s5)
    0x03, 0xc0, 0xc0, 0x25,                             //move    $t8, $fp
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x03, 0x80, 0x10, 0x25,                             //move    $v0, $gp
    0x02, 0x40, 0xe0, 0x25,                             //move    $gp, $s2
    0x26, 0xb5, 0x00, 0x20,                             //addiu   $s5, 0x20
    0x26, 0x94, 0x00, 0x01,                             //addiu   $s4, 1
    0x2e, 0x81, 0x00, 0x10,                             //sltiu   $1, $s4, 0x10
    0x54, 0x20, 0xff, 0xe3,                             //bnezl   $1, XXXX
    0x8e, 0xb2, 0x00, 0x10,                             //lw      $s2, 0x10($s5)
    0x8f, 0xd9, 0x00, 0x00,                             //lw      $t9, 0x00($fp)
    0x53, 0x20, 0x00, 0x03,                             //beqzl   $t9, XXXX
    0x8f, 0xbf, 0x00, 0x24,                             //lw      $ra, 0x24($sp)
    0xaf, 0xd6, 0x00, 0x00,                             //sw      $s6, 0x00($fp)
    0x8f, 0xbf, 0x00, 0x24,                             //lw      $ra, 0x24($sp)
};

//fw_ffindfirst
byte pattern_fw_ffindfirst[] = {
    0x27, 0xbd, 0xff, 0xc0,                             //addiu   $sp, -0x40
    0xaf, 0xbe, 0x00, 0x3c,                             //sw      $fp, 0x3c($sp)
    0xaf, 0xb3, 0x00, 0x2c,                             //sw      $s3, 0x2c($sp)
    0xaf, 0xb4, 0x00, 0x30,                             //sw      $s4, 0x30($sp)
    0xaf, 0xb5, 0x00, 0x34,                             //sw      $s5, 0x34($sp)
    0xaf, 0xb6, 0x00, 0x38,                             //sw      $s6, 0x38($sp)
    0xaf, 0xbf, 0x00, 0x28,                             //sw      $ra, 0x28($sp)
    0x00, 0x80, 0xf0, 0x25,                             //move    $fp, $a0
    0x8f, 0x84, 0xFF, 0xFF,                             //lw      $a0, -0x6168($gp)         ; <---
    0x00, 0xa0, 0xb0, 0x25,                             //move    $s6, $a1
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX
    0x27, 0xa5, 0x00, 0x24,                             //addiu   $a1, $sp, 0x24
    0x30, 0x42, 0xff, 0xff,                             //andi    $v0, 0xffff
    0x10, 0x40, 0x00, 0x03,                             //beqz    $v0, XXXX
    0x3c, 0x13, 0xFF, 0xFF,                             //lui     $s3, XXXX
    0x10, 0x00, 0x00, 0x4c,                             //b       XXXX
};

//fw_tap_load
byte pattern_fw_tap_load[] = {
    0x27, 0xbd, 0xff, 0xb0,                             //addiu   $sp, -0x50
    0xaf, 0xb1, 0x00, 0x34,                             //sw      $s1, pTAP_GetState($sp)
    0xaf, 0xb2, 0x00, 0x38,                             //sw      $s2, pTAP_ExitNormal($sp)
    0xaf, 0xb3, 0x00, 0x3c,                             //sw      $s3, pTAP_EnterNormal($sp)
    0xaf, 0xb4, 0x00, 0x40,                             //sw      $s4, 0x40($sp)
    0xaf, 0xb5, 0x00, 0x44,                             //sw      $s5, 0x44($sp)
    0xaf, 0xb6, 0x00, 0x48,                             //sw      $s6, pTAP_Exit($sp)
    0xaf, 0xbe, 0x00, 0x4c,                             //sw      $fp, pTAP_Kbhit($sp)
    0xaf, 0xbf, 0x00, 0x30,                             //sw      $ra, 0x30($sp)
    0x80, 0x98, 0x00, 0x00,                             //lb      $t8, 0($a0)
    0x24, 0xFF, 0x00, 0xFF,                             //li      $XX, XX
    0x13, 0x00, 0x00, 0x0d,                             //beqz    $t8, XXXX
    0x00, 0x00, 0xFF, 0x25,                             //move    $XX, $0
    0x00, 0x80, 0xFF, 0x25,                             //move    $XX, $a0
    0x82, 0xFF, 0x00, 0x00,                             //lb      $XX, 0($XX)
    0x24, 0x18, 0x00, 0x0d,                             //li      $t8, 0xd
};

//fw_tap_load_data1
byte pattern_fw_tap_load_data1[] = {
    0x27, 0xbd, 0xff, 0xe8,                             //addiu   $sp, -0x18
    0xaf, 0xbf, 0x00, 0x14,                             //sw      $ra, 0x14($sp)
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX
    0x27, 0x84, 0xFF, 0xFF,                             //addiu   $a0, $gp, 0x875c
    0xa3, 0x82, 0xFF, 0xFF,                             //sb      $v0, -0x78a4($gp)
    0x3c, 0x05, 0xFF, 0xFF,                             //la      $a1, XXXXXXXX
    0x24, 0xa5, 0xFF, 0xFF,                             //
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX
    0x34, 0x04, 0xe3, 0x00,                             //li      $a0, 0xe300
    0x83, 0x8e, 0xFF, 0xFF,                             //lb      $t6, -0x789c($gp)
    0x11, 0xc0, 0x00, 0x0c,                             //beqz    $t6, XXXX
    0x24, 0x04, 0x00, 0xff,                             //li      $a0, 0xff
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX
    0x8f, 0x84, 0xFF, 0xFF,                             //lw      $a0, -0x7af4($gp)
    0x27, 0x84, 0xFF, 0xFF,                             //addiu   $a0, $gp, 0x850c
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX
    0x8f, 0x85, 0xFF, 0xFF,                             //lw      $a1, -0x7b00($gp)
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX
    0x8f, 0x84, 0xFF, 0xFF,                             //lw      $a0, -0x7af4($gp)
    0x3c, 0x04, 0xFF, 0xFF,                             //lui     $a0, XXXX
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     load_tap
    0x24, 0x84, 0xFF, 0xFF,                             //li      $a0, XXXXXXXX
    0x24, 0x04, 0x00, 0xff,                             //li      $a0, 0xff
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     enter_normal
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX
    0x8f, 0x84, 0xFF, 0xFF,                             //lw      $a0, -0x7b04($gp)
    0x8f, 0xbf, 0x00, 0x14,                             //lw      $ra, 0x14($sp)
    0x03, 0xe0, 0x00, 0x08,                             //jr      $ra
    0x27, 0xbd, 0x00, 0x18,                             //addiu   $sp, 0x18
};

//
////fw_led_patch
//byte pattern_fw_led_patch[] = {
//};

//
//fw_num_ext_functions
byte pattern_fw_TAP_GetSystemProc[] = {
    0x3c, 0x05, 0xFF, 0xFF,                             //la      $a1, TapIdent
    0x24, 0xa5, 0xFF, 0xFF,                             //
    0x8c, 0xaf, 0x00, 0x00,                             //lw      $t7, 0($a1)
    0x3c, 0x03, 0xFF, 0xFF,                             //la      $v1, TapTable
    0x24, 0x63, 0xFF, 0xFF,                             //
    0x00, 0x0f, 0x79, 0x40,                             //sll     $t7, 5
    0x00, 0x6f, 0x78, 0x21,                             //addu    $t7, $v1, $t7
    0x8d, 0xe6, 0x00, 0x00,                             //lw      $a2, 0($t7)
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x03, 0x80, 0x10, 0x25,                             //move    $v0, $gp
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0xc0, 0xe0, 0x25,                             //move    $gp, $a2
    0x28, 0x81, 0xFF, 0xFF,                             //slti    $1, $a0, 0x15
    0x14, 0x20, 0x00, 0x03,                             //bnez    $1, XXXX
    0x3c, 0x19, 0xFF, 0xFF,                             //lui     $t9, XXXX
    0x10, 0x00, 0x00, 0x05,                             //b       XXXX
    0x00, 0x00, 0x30, 0x25,                             //move    $a2, $0
    0x27, 0x39, 0xFF, 0xFF,                             //la      $t9, XXXXXXXX
};

//fw_enable_tshift
byte pattern_fw_enable_tshift[] = {
    0x27, 0xbd, 0xff, 0xe8,                             //addiu   $sp, -0x18
    0xaf, 0xbf, 0x00, 0x14,                             //sw      $ra, 0x18+var_4($sp)
    0x83, 0x99, 0xFF, 0xFF,                             //lb      $t9, -0x7B44($gp)
    0x24, 0x18, 0x00, 0x01,                             //li      $t8, 1
    0x57, 0x38, 0x00, 0x03,                             //bnel    $t9, $t8, loc_80143120
    0x97, 0x99, 0xFF, 0xFF,                             //lhu     $t9, -0x5EBC($gp)
    0x10, 0x00, 0x00, 0x8c,                             //b       loc_8014334C
    0x00, 0x00, 0x10, 0x25,                             //move    $v0, $0
    0x57, 0x20, 0x00, 0x07,                             //bnezl   $t9, loc_80143140
    0x83, 0x99, 0xFF, 0xFF,                             //lb      $t9, -0x79FF($gp)
    0x97, 0x99, 0xFF, 0xFF,                             //lhu     $t9, -0x5EBA($gp)
    0x57, 0x20, 0x00, 0x04,                             //bnezl   $t9, loc_80143140
    0x83, 0x99, 0xFF, 0xFF,                             //lb      $t9, -0x79FF($gp)
    0x10, 0x00, 0x00, 0x85,                             //b       loc_8014334C
    0x00, 0x00, 0x10, 0x25,                             //move    $v0, $0
    0x83, 0x99, 0xFF, 0xFF,                             //lb      $t9, -0x79FF($gp)
};


//fw_stop_tshift
byte pattern_fw_stop_tshift[] = {
    0x27, 0xbd, 0xff, 0xd0,                             //addiu   $sp, -0x30
    0xaf, 0xbf, 0x00, 0x14,                             //sw      $ra, 0x30+var_1C($sp)
    0xaf, 0xb2, 0x00, 0x18,                             //sw      $s2, 0x30+var_18($sp)
    0xaf, 0xb3, 0x00, 0x1c,                             //sw      $s3, 0x30+var_14($sp)
    0xaf, 0xb4, 0x00, 0x20,                             //sw      $s4, 0x30+var_10($sp)
    0xaf, 0xb5, 0x00, 0x24,                             //sw      $s5, 0x30+var_C($sp)
    0xaf, 0xb6, 0x00, 0x28,                             //sw      $s6, 0x30+var_8($sp)
    0xaf, 0xbe, 0x00, 0x2c,                             //sw      $fp, 0x30+var_4($sp)
    0x00, 0x80, 0x90, 0x25,                             //move    $s2, $a0                      ; flag
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     kill_task3
    0x27, 0x84, 0xFF, 0xFF,                             //addiu   $a0, $gp, 0x87A8
    0xa3, 0x82, 0xFF, 0xFF,                             //sb      $v0, -0x7858($gp)
    0x93, 0xFF, 0xFF, 0xFF,                             //lbu     $s6, -0x7860($gp)             ; rec_slot
    0x24, 0x0e, 0x00, 0xFF,                             //li      $t6, 0x88
    0x3c, 0x14, 0xFF, 0xFF,                             //lui     $s4, 0x81CA
    0xFF, 0xce, 0x00, 0x18,                             //mult    $s6, $t6
};

//for early aus-416
byte pattern_fw_stop_tshift2[] = {
    0x27, 0xbd, 0xff, 0xd8,                             //addiu   $sp, -0x28
    0xaf, 0xbf, 0x00, 0x14,                             //sw      $ra, 0x28+var_14($sp)
    0xaf, 0xb4, 0x00, 0x18,                             //sw      $s4, 0x28+var_10($sp)
    0xaf, 0xb5, 0x00, 0x1c,                             //sw      $s5, 0x28+var_C($sp)
    0xaf, 0xb6, 0x00, 0x20,                             //sw      $s6, 0x28+var_8($sp)
    0xaf, 0xbe, 0x00, 0x24,                             //sw      $fp, 0x28+var_4($sp)
    0x00, 0x80, 0xa0, 0x25,                             //move    $s4, $a0
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     kill_task3
    0x27, 0x84, 0xFF, 0xFF,                             //addiu   $a0, $gp, 0x8740
    0xa3, 0x82, 0xFF, 0xFF,                             //sb      $v0, -0x78C0($gp)
    0x93, 0x9e, 0xFF, 0xFF,                             //lbu     $fp, -0x78C8($gp)
    0x24, 0x0e, 0xFF, 0xFF,                             //li      $t6, 0x88
    0x3c, 0x15, 0xFF, 0xFF,                             //lui     $s5, 0x819A
    0x03, 0xce, 0x00, 0x18,                             //mult    $fp, $t6
    0x26, 0xb5, 0xFF, 0xFF,                             //li      $s5, 0x819A31C8
    0x00, 0x00, 0x70, 0x12,                             //mflo    $t6
};


//fw_mheg_disable
byte pattern_fw_mheg_disable[] = {
    0x27, 0xbd, 0xff, 0xFF,                             //addiu   $sp, -0x38
    0xaf, 0xbf, 0x00, 0xFF,                             //sw      $ra, 0x38+var_4($sp)
    0x83, 0x99, 0xFF, 0xFF,                             //lb      $t9, -0x74C9($gp)
    0x53, 0x20, 0x00, 0xFF,                             //beqzl   $t9, loc_801B6C40
    0xa3, 0x80, 0xFF, 0xFF,                             //sb      $0, -0x74C8($gp)
    0x24, 0x19, 0x00, 0xFF,                             //li      $t9, 0xE
    0xaf, 0xb9, 0x00, 0xFF,                             //sw      $t9, 0x38+var_14($sp)
    0x27, 0xa5, 0x00, 0xFF,                             //addiu   $a1, $sp, 0x38+var_14
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     ioctl
    0x24, 0x04, 0x00, 0x13,                             //li      $a0, 0x13
    0x3c, 0x04, 0xFF, 0xFF,                             //li      $a0, 0x81D4A184
    0x24, 0x84, 0xFF, 0xFF,                             //
    0x24, 0x05, 0x00, 0xff,                             //li      $a1, 0xFF
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     memset3
    0x24, 0x06, 0x00, 0x30,                             //li      $a2, 0x30
};

//fw_ioctl
byte pattern_fw_ioctl[] = {
    0x27, 0xbd, 0xff, 0xd0,                             //addiu   $sp, -0x30
    0xaf, 0xb2, 0x00, 0x18,                             //sw      $s2, 0x30+var_18($sp)
    0xaf, 0xb3, 0x00, 0x1c,                             //sw      $s3, 0x30+var_14($sp)
    0xaf, 0xb4, 0x00, 0x20,                             //sw      $s4, 0x30+var_10($sp)
    0xaf, 0xb5, 0x00, 0x24,                             //sw      $s5, 0x30+var_C($sp)
    0xaf, 0xb6, 0x00, 0x28,                             //sw      $s6, 0x30+var_8($sp)
    0xaf, 0xbe, 0x00, 0x2c,                             //sw      $fp, 0x30+var_4($sp)
    0xaf, 0xbf, 0x00, 0x14,                             //sw      $ra, 0x30+var_1C($sp)
    0x00, 0x80, 0xf0, 0x25,                             //move    $fp, $a0
    0x00, 0xa0, 0xb0, 0x25,                             //move    $s6, $a1
    0x2f, 0xc1, 0x00, 0x14,                             //sltiu   $1, $fp, 0x14
    0x14, 0x20, 0x00, 0x04,                             //bnez    $1, loc_80000CA8
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x24, 0x02, 0x00, 0x90,                             //li      $v0, 0x90
    0x10, 0x00, 0x00, 0xFF,                             //b       loc_80000E24
    0x00, 0x00, 0x00, 0x00,                             //nop

    0x24, 0x12, 0x00, 0x02,                             //li      $s2, 2
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     os_lock
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x40, 0xa8, 0x25,                             //move    $s5, $v0
    0x3c, 0x18, 0xFF, 0xFF,                             //li      $t8, 0x8069F730
    0x27, 0x18, 0xFF, 0xFF,                             //
    0x00, 0x1e, 0x78, 0x80,                             //sll     $t7, $fp, 2
    0x03, 0x0f, 0xc0, 0x21,                             //addu    $t8, $t7
    0x8f, 0x14, 0x00, 0x00,                             //lw      $s4, 0($t8)
    0x26, 0x93, 0x00, 0x04,                             //addiu   $s3, $s4, 4
    0x3c, 0x0f, 0xFF, 0xFF,                             //la      $t7, unk_802C8B78
    0x25, 0xef, 0xFF, 0xFF,                             //
    0x00, 0x1e, 0x70, 0x80,                             //sll     $t6, $fp, 2
    0x01, 0xee, 0x78, 0x21,                             //addu    $t7, $t6
    0x8d, 0xef, 0x00, 0x00,                             //lw      $t7, 0($t7)
    0x02, 0x6f, 0x08, 0x2b,                             //sltu    $1, $s3, $t7
};

//epg_stick_patch     - 80120978
byte pattern_fw_epgstck_patch[] = {
    0x00, 0x18, 0xc7, 0xc2,                             //srl     $t8, 31          `    ; --> 24 18 00 01   li $t8, 1
    0x57, 0x00, 0x00, 0x0f,                             //bnezl   $t8, loc_801209D8     ; always take to avoid channel change on epg
    0x8f, 0xbf, 0x00, 0x20,                             //lw      $ra, 0x28+var_8($sp)
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     kill_task3
    0x27, 0x84, 0xFF, 0xFF,                             //addiu   $a0, $gp, 0x86C0
    0xa3, 0x82, 0xFF, 0xFF,                             //sb      $v0, -0x7940($gp)
    0x87, 0x98, 0xFF, 0xFF,                             //lh      $t8, -0x666C($gp)
    0x3c, 0x06, 0xFF, 0xFF,                             //lui     $a2, 0x8012
    0x00, 0x18, 0xc0, 0x80,                             //sll     $t8, 2
    0x03, 0xd8, 0xc0, 0x21,                             //addu    $t8, $fp, $t8
    0x8f, 0x18, 0x00, 0x00,                             //lw      $t8, 0($t8)
    0x27, 0x84, 0xFF, 0xFF,                             //addiu   $a0, $gp, 0x86C0
    0x97, 0x07, 0x00, 0x08,                             //lhu     $a3, 8($t8)
    0x24, 0xc6, 0xFF, 0xFF,                             //la      $a2, 0x8011E52C
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     register_task_4
    0x24, 0x05, 0x00, 0x32,                             //li      $a1, 0x32
};

//fw_channel_scale()
byte pattern_fw_channel_scale[] = {
    0x27, 0xbd, 0xff, 0xd0,                             //addiu   $sp, -0x30
    0xaf, 0xbe, 0x00, 0x2c,                             //sw      $fp, 0x30+var_4($sp)
    0x00, 0xc0, 0x48, 0x25,                             //move    $t1, $a2         # y
    0x8f, 0xbe, 0x00, 0x40,                             //lw      $fp, 0x30+arg_10($sp)  # h
    0x8f, 0xa6, 0x00, 0x44,                             //lw      $a2, 0x30+arg_14($sp)  # mode
    0xaf, 0xbf, 0x00, 0x28,                             //sw      $ra, 0x30+var_8($sp)
    0x10, 0xe0, 0x00, 0x03,                             //beqz    $a3, loc_80105110  # w
    0x00, 0xe0, 0x50, 0x25,                             //move    $t2, $a3         # w
    0x17, 0xc0, 0x00, 0x04,                             //bnez    $fp, loc_8010511c
    0x00, 0x0a, 0xc8, 0x40,                             //sll     $t9, $t2, 1
    0x24, 0x0a, 0x02, 0xd0,                             //li      $t2, 0x2d0       # 720
    0x24, 0x1e, 0x02, 0x40,                             //li      $fp, 0x240       # 576
    0x00, 0x0a, 0xc8, 0x40,                             //sll     $t9, $t2, 1
    0x24, 0x18, 0x00, 0x64,                             //li      $t8, 0x64        # 100
    0x03, 0x38, 0x00, 0x1a,                             //div     $t9, $t8
    0x00, 0x00, 0x08, 0x12,                             //mflo    $1
};
byte pattern_fw_channel_scale2[] = {
    0x27, 0xbd, 0xff, 0xd0,                             //addiu   $sp, -0x30
    0x8f, 0xa8, 0x00, 0x40,                             //lw      $t0, 0x30+arg_10($sp)
    0x8f, 0xaa, 0x00, 0x44,                             //lw      $t2, 0x30+arg_14($sp)
    0xaf, 0xb6, 0x00, 0x28,                             //sw      $s6, 0x30+var_8($sp)
    0xaf, 0xbe, 0x00, 0x2c,                             //sw      $fp, 0x30+var_4($sp)
    0x10, 0xe0, 0x00, 0x03,                             //beqz    $a3, loc_80104E6C
    0xaf, 0xbf, 0x00, 0x24,                             //sw      $ra, 0x30+var_C($sp)
    0x15, 0x00, 0x00, 0x04,                             //bnez    $t0, loc_80104E78
    0x00, 0x05, 0xc8, 0x42,                             //srl     $t9, $a1, 1
    0x24, 0x07, 0x02, 0xd0,                             //li      $a3, 0x2D0       # 720
    0x24, 0x08, 0x02, 0x40,                             //li      $t0, 0x240       # 576
    0x00, 0x05, 0xc8, 0x42,                             //srl     $t9, $a1, 1      # x
    0x04, 0xa1, 0x00, 0x04,                             //bgez    $a1, loc_80104E8C
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x05, 0xc8, 0x23,                             //negu    $t9, $a1
    0x00, 0x19, 0xc8, 0x42,                             //srl     $t9, 1
};

//fw_set_wss()
byte pattern_fw_set_wss[] = {
    0x27, 0xbd, 0xff, 0xe0,                             //addiu   $sp, -0x20
    0xaf, 0xb5, 0x00, 0x14,                             //sw      $s5, 0x20+var_c($sp)
    0xaf, 0xb6, 0x00, 0x18,                             //sw      $s6, 0x20+var_8($sp)
    0xaf, 0xbe, 0x00, 0x1c,                             //sw      $fp, 0x20+var_4($sp)
    0x00, 0x80, 0xf0, 0x25,                             //move    $fp, $a0
    0x00, 0xa0, 0xb0, 0x25,                             //move    $s6, $a1
    0x00, 0x00, 0xa8, 0x25,                             //move    $s5, $0
    0x93, 0x99, 0xFF, 0xFF,                             //lbu     $t9, -0x7ea4($gp)
    0x17, 0x20, 0x00, 0x04,                             //bnez    $t9, loc_80050a34
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x24, 0x02, 0x00, 0x01,                             //li      $v0, 1
    0x10, 0x00, 0x00, 0x18,                             //b       loc_80050a90
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x8f, 0x99, 0xFF, 0xFF,                             //lw      $t9, -0x7ea0($gp)
    0x24, 0x18, 0x00, 0x02,                             //li      $t8, 2
    0x17, 0x38, 0x00, 0x0b,                             //bne     $t9, $t8, loc_80050a6c
};

byte pattern_fw_set_wss2[] = {
    0x27, 0xbd, 0xff, 0xe0,                             //addiu   $sp, -0x20
    0xaf, 0xb5, 0x00, 0x14,                             //sw      $s5, 0x20+var_C($sp)
    0xaf, 0xb6, 0x00, 0x18,                             //sw      $s6, 0x20+var_8($sp)
    0xaf, 0xbe, 0x00, 0x1c,                             //sw      $fp, 0x20+var_4($sp)
    0xaf, 0xbf, 0x00, 0x10,                             //sw      $ra, 0x20+var_10($sp)
    0x00, 0x80, 0xf0, 0x25,                             //move    $fp, $a0
    0x00, 0xa0, 0xb0, 0x25,                             //move    $s6, $a1
    0x00, 0x00, 0xa8, 0x25,                             //move    $s5, $0
    0x93, 0x99, 0xFF, 0xFF,                             //lbu     $t9, -0x7ea0($gp)
    0x17, 0x20, 0x00, 0x08,                             //bnez    $t9, loc_80050A34
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x3c, 0x04, 0xFF, 0xFF,                             //la      $a0, aWssInitialised  # "WSS initialised false....\n"
    0x24, 0x84, 0xFF, 0xFF,                             //
    0x0c, 0xFF, 0xFF, 0xFF,                             //jal     rs232_printf
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x24, 0x02, 0x00, 0x01,                             //li      $v0, 1
};


//fw_pip_position_offset
//ETC structure	structure	$gp - 0x5f24
byte pattern_TAP_GetSignalLevel[] = {
    0x27, 0xbd, 0xff, 0xe0,                             //addiu   $sp, -0x20
    0xaf, 0xbe, 0x00, 0x1c,                             //sw      $fp, 0x20+var_4($sp)
    0x3c, 0x1e, 0xFF, 0xFF,                             //lui     $fp, 0x8017
    0xaf, 0xbf, 0x00, 0x14,                             //sw      $ra, 0x20+var_C($sp)
    0x27, 0xde, 0xFF, 0xFF,                             //la      $fp, TapIdent
    0xaf, 0xb6, 0x00, 0x18,                             //sw      $s6, 0x20+var_8($sp)
    0x8f, 0xcf, 0x00, 0x00,                             //lw      $t7, 0x20+var_20($fp)
    0x3c, 0x16, 0xFF, 0xFF,                             //la      $s6, TapTable
    0x26, 0xd6, 0xFF, 0xFF,                             //la      $s6, TapTable
    0x00, 0x0f, 0x79, 0x40,                             //sll     $t7, 5
    0x02, 0xcf, 0x78, 0x21,                             //addu    $t7, $s6, $t7
    0x8d, 0xe3, 0x00, 0x00,                             //lw      $v1, 0($t7)
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x03, 0x80, 0x10, 0x25,                             //move    $v0, $gp
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x60, 0xe0, 0x25,                             //move    $gp, $v1
    0x8f, 0x99, 0xFF, 0xFF,                             //lw      $t9, -0x5ED4($gp)       <--- ETC structure
    0x27, 0xa5, 0x00, 0x10,                             //addiu   $a1, $sp, 0x20+var_10
};

//fw_afd
byte pattern_fw_afd[] = {
    0x27, 0xbd, 0xff, 0xd0,                             //addiu   $sp, -0x30
    0xaf, 0xb6, 0x00, 0x28,                             //sw      $s6, 0x30+var_8($sp)
    0xaf, 0xbe, 0x00, 0x2c,                             //sw      $fp, 0x30+var_4($sp)
    0xaf, 0xbf, 0x00, 0x24,                             //sw      $ra, 0x30+var_C($sp)
    0x00, 0x80, 0xf0, 0x25,                             //move    $fp, $a0
    0x00, 0xa0, 0xb0, 0x25,                             //move    $s6, $a1
    0x32, 0xd6, 0x00, 0xff,                             //andi    $s6, 0xFF
    0x24, 0x19, 0x00, 0x01,                             //li      $t9, 1
    0x16, 0xd9, 0x01, 0xa5,                             //bne     $s6, $t9, loc_80083308
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x93, 0x99, 0xFF, 0xFF,                             //lbu     $t9, -0x6354($gp)  # main_afd
    0x93, 0x98, 0xFF, 0xFF,                             //lbu     $t8, -0x7CB4($gp)  # main_afd
    0x13, 0x38, 0x02, 0x7a,                             //beq     $t9, $t8, loc_8008366C
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x93, 0x99, 0xFF, 0xFF,                             //lbu     $t9, -0x6354($gp)  # main_afd
    0xa3, 0x99, 0xFF, 0xFF,                             //sb      $t9, -0x7CB4($gp)  # main_afd
};

//fw_sequence_header
byte pattern_fw_sequence_header[] = {
    0x32, 0x10, 0x00, 0xFF,                             //andi    $s0, 0xFF        # sequence_header
    0x02, 0x00, 0x20, 0x25,                             //move    $a0, $s0
    0x0C, 0xFF, 0xFF, 0xFF,                             //jal     XXXXXX
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x00, 0x40, 0x98, 0x25,                             //move    $s3, $v0
    0x24, 0x19, 0x00, 0x01,                             //li      $t9, 1
    0x16, 0x79, 0x00, 0x09,                             //bne     $s3, $t9, loc_800D1D24
    0x00, 0x00, 0x00, 0x00,                             //nop
    0x8F, 0x99, 0xFF, 0xFF,                             //lw      $t9, -0x636C($gp)     <---
    0x8F, 0x98, 0xFF, 0xFF,                             //lw      $t8, -0x636C($gp)
    0x8F, 0x18, 0x00, 0x10,                             //lw      $t8, 0x10($t8)
    0x3C, 0x0F, 0x10, 0x00,                             //lui     $t7, 0x1000
    0x03, 0x0F, 0xC0, 0x25,                             //or      $t8, $t7
    0xAF, 0x38, 0x00, 0x10,                             //sw      $t8, 0x10($t9)
    0x10, 0x00, 0x00, 0x71,                             //b       loc_800D1EE4
    0x00, 0x00, 0x00, 0x00,                             //nop
};

//fw_get_genre_str
byte pattern_fw_get_genre_str[] = {
    0x27, 0xbd, 0xff, 0xe8,                             //addiu   $sp, -0x18
    0xaf, 0xbf, 0x00, 0x14,                             //sw      $ra, 0x18+var_4($sp)
    0x90, 0xb9, 0x00, 0x40,                             //lbu     $t9, 0x40($a1)   # nibble1
    0x2b, 0x21, 0x00, 0x0b,                             //slti    $1, $t9, 0xB
    0x54, 0x20, 0x00, 0x03,                             //bnezl   $1, loc_8012B1BC
    0x90, 0xb8, 0x00, 0x40,                             //lbu     $t8, 0x40($a1)   # if (nibble1 > 0xb) nibble1 = 0
    0xa0, 0xa0, 0x00, 0x40,                             //sb      $0, 0x40($a1)
    0x90, 0xb8, 0x00, 0x40,                             //lbu     $t8, 0x40($a1)
    0x24, 0x0f, 0x00, 0x4c,                             //li      $t7, 0x4C
    0x3c, 0x19, 0xFF, 0xFF,                             //lui     $t9, 0x802F
    0x03, 0x0f, 0x00, 0x18,                             //mult    $t8, $t7
    0x83, 0x8e, 0xFF, 0xFF,                             //lb      $t6, -0x6188($gp)  # language
    0x27, 0x39, 0xFF, 0xFF,                             //la      $t9, off_802E88D4  # genre
    0x00, 0x00, 0xc0, 0x12,                             //mflo    $t8
    0x03, 0x38, 0xc8, 0x21,                             //addu    $t9, $t8
    0x00, 0x0e, 0x70, 0x80,                             //sll     $t6, 2
};

//fw_epg_highlighted_event     80121B64
byte pattern_fw_epg_highlighted_event[] = {
    0x3c, 0x18, 0xFF, 0xFF,                             //lui     $t8, 0x802E
    0x27, 0xbd, 0xff, 0xe8,                             //addiu   $sp, -0x18
    0x27, 0x18, 0xFF, 0xFF,                             //la      $t8, unk_802DF1D4
    0x00, 0x04, 0x78, 0x80,                             //sll     $t7, $a0, 2
    0xaf, 0xbe, 0x00, 0x14,                             //sw      $fp, 0x18+var_4($sp)
    0xaf, 0xbf, 0x00, 0x10,                             //sw      $ra, 0x18+var_8($sp)
    0x03, 0x0f, 0xc0, 0x21,                             //addu    $t8, $t7
    0x8f, 0x18, 0x00, 0x00,                             //lw      $t8, 0($t8)
    0x00, 0x05, 0xc8, 0x80,                             //sll     $t9, $a1, 2
    0x3c, 0x1e, 0xFF, 0xFF,                             //lui     $fp, 0x81B5             <---
    0x03, 0x19, 0xc0, 0x21,                             //addu    $t8, $t9
    0x27, 0xde, 0xFF, 0xFF,                             //li      $fp, 0x81B52B7C         <---
    0x8f, 0x0c, 0x00, 0x00,                             //lw      $t4, 0($t8)
    0x03, 0xc0, 0x58, 0x25,                             //move    $t3, $fp
    0x34, 0x0a, 0x00, 0x08,                             //li      $t2, 8
    0x8d, 0x8e, 0x00, 0x00,                             //lw      $t6, 0($t4)
};
//fw_epg_refresh
byte pattern_fw_epg_refresh[] = {
    0x27, 0xbd, 0xff, 0xb8,                             //addiu   $sp, -0x48
    0xaf, 0xb4, 0x00, 0x38,                             //sw      $s4, 0x48+var_10($sp)
    0xaf, 0xb5, 0x00, 0x3c,                             //sw      $s5, 0x48+var_C($sp)
    0xaf, 0xb6, 0x00, 0x40,                             //sw      $s6, 0x48+var_8($sp)
    0xaf, 0xbe, 0x00, 0x44,                             //sw      $fp, 0x48+var_4($sp)
    0xaf, 0xbf, 0x00, 0x34,                             //sw      $ra, 0x48+var_14($sp)
    0x83, 0x99, 0xFF, 0xFF,                             //lb      $t9, -0x62AC($gp)         <----
    0x24, 0x18, 0x00, 0x18,                             //li      $t8, 0x18
    0x17, 0x38, 0xFF, 0xFF,                             //bne     $t9, $t8, loc_80128724
    0x00, 0x00, 0xa8, 0x25,                             //move    $s5, $0
    0x24, 0x19, 0x00, 0x01,                             //li      $t9, 1
    0xa3, 0x99, 0xFF, 0xFF,                             //sb      $t9, -0x7804($gp)
};


int search_function(char *fn_name, byte *pattern, int pattern_length, byte *base, dword start, dword end, dword *result)
{
    int found;
    int i;

    printf("\nsearching for %s ... ", fn_name);
    found = search_pattern(pattern, pattern_length, base, start, end, result);
    if (found < 0) {
        printf("failed, err=%d", found);
        return -1;
    }
    else
    if (found == 0) {
        printf("not found");
        return 0;
    }
    else
    if (found == 1) {
        result[0] = 0x80000000+result[0];
        printf("found at %08x", result[0]);
        return result[0];
    }
    else {
        printf("found %d occurances", found);
        for (i=0; i<found; i++) {
            result[i] = 0x80000000+result[i];
            printf("\n  found %d at %08x", i+1, result[i]);
            return result[0];
        }
    }
    return found;
}

int search_variable(char *var_name, byte *pattern, int msb, int lsb, int pattern_length, byte *base, dword start, dword end, dword *result)
{
    int found;
    int i;
    dword var;

    printf("\nsearching for %s ... ", var_name);
    found = search_pattern(pattern, pattern_length, base, start, end, result);
    if (found < 0) {
        printf("failed, err=%d", found);
        return -1;
    }
    else
    if (found == 0) {
        printf("not found");
        return 0;
    }
    else {
        printf("found %d occurances", found);
//        for (i=0; i<found; i++) {
//            printf("\n  found %d at %08x", i+1, result[i]);
//        }
        //16 bit lsb is signed ...
        var = (to_be(*(word *)(base+result[0]+msb)) << 16) + (short)(to_be(*(word *)(base+result[0]+lsb)));
        if (msb == lsb) var &= 0xffff;
        printf(", variable = %08x", var);
        return var;
    }
    return found;
}

int search_gpvar(char *var_name, byte *pattern, dword gp, int offset, int pattern_length, byte *base, dword start, dword end, dword *result)
{
    int found;
    int i;
    dword var;

    printf("\nsearching for %s ... ", var_name);
    found = search_pattern(pattern, pattern_length, base, start, end, result);
    if (found < 0) {
        printf("failed, err=%d", found);
        return -1;
    }
    else
    if (found == 0) {
        printf("not found");
        return 0;
    }
    else {
        printf("found %d occurances", found);
//         for (i=0; i<found; i++) {
//             printf("\n  found %d at %08x", i+1, result[i]);
//         }
        //16 bit offset is signed ...
        var = gp + (short)(to_be(*(word *)(base+result[0]+offset)));
        printf(", variable = %08x", var);
        return var;
    }
    return var;
}

int get_jump(char *fn_name, dword pattern, word start, dword end, byte *base, dword *result)
{
    dword found;
    dword offset;

    printf("\nsearching for %s ... ", fn_name);
    found = 0;
    for (offset=(pattern-0x80000000+start)&~3; offset < (pattern-0x80000000+end); offset +=4) {
        if (base[offset] == 0x0c) {   //jal
            result[0] = base[offset+1];
            result[0] <<= 8;
            result[0] += base[offset+2];
            result[0] <<= 8;
            result[0] += base[offset+3];
            result[0] *= 4;
            result[0] += 0x80000000;
            found = result[0];
            printf("found at %08x", found);
            return found;
        }
    }

    printf("not found");
    return 0;
}


type_fw_addresses fw;

int test(byte *base, int i)
{
    dword result[MAX_FOUND];

    fw.fw_appl_version[i]         = search_variable("fw_appl_version        ", pattern_fw_appl_version,                 0x02, 0x02, 0x40, base, 0x00000000, 0x00250000, result);
    if (fw.fw_appl_version[i] == 0)
        fw.fw_appl_version[i]     = search_variable("fw_appl_version        ", pattern_fw_appl_version2,                0x16, 0x16, 0x40, base, 0x00000000, 0x00250000, result);
    if (fw.fw_appl_version[i] == 0)
        fw.fw_appl_version[i]     = search_variable("fw_appl_version        ", pattern_fw_appl_version3,                0x02, 0x02, 0x40, base, 0x00000000, 0x00250000, result);
    fw.fw_gp[i]                   = search_variable("fw_gp                  ", pattern_fw_gp,                           0x26, 0x2a, 0x40, base, 0x00000000, 0x00250000, result);

    fw.fw_tap_table[i]            = search_variable("fw_tap_table           ", pattern_TAP_KbHit,                       0x1e, 0x22, 0x40, base, 0x00000000, 0x00250000, result);
    fw.fw_tap_ident[i]            = search_variable("fw_tap_ident           ", pattern_TAP_KbHit,                       0x0a, 0x12, 0x40, base, 0x00000000, 0x00250000, result);

    fw.fw_os_free[i]              = search_function("fw_os_free             ", pattern_fw_os_free,                                  0x40, base, 0x00000000, 0x00250000, result);
    fw.fw_add_task_queue[i]       = search_function("fw_add_task_queue      ", pattern_fw_add_task_queue,                           0x40, base, 0x00000000, 0x00250000, result);

    fw.fw_shutdown_irq_disable[i] = search_function("fw_shutdown_irq_disable", pattern_fw_shutdown_irq_disable,                     0x34, base, 0x00000000, 0x00250000, result);
    fw.fw_shutdown_irq_enable[i]  = search_function("fw_shutdown_irq_enable ", pattern_fw_shutdown_irq_enable,                      0x34, base, 0x00000000, 0x00250000, result);
    fw.fw_shutdown_task[i]        = search_function("fw_shutdown_task       ", pattern_fw_shutdown_task,                            0x40, base, 0x00000000, 0x00250000, result);

    fw.fw_process_tdt[i]          = search_function("fw_process_tdt         ", pattern_process_tdt,                                 0x70, base, 0x00000000, 0x00250000, result);
    fw.fw_process_tot[i]          = search_function("fw_process_tot         ", pattern_process_tot,                                 0x40, base, 0x00000000, 0x00250000, result);
    if (fw.fw_process_tot[i] == 0) {
        fw.fw_process_tot[i]      = search_function("fw_process_tot         ", pattern_process_tot2,                                0x40, base, 0x00000000, 0x00250000, result);
    }
    if (fw.fw_process_tot[i]) {
        fw.fw_DbgSi_LoadDmx[i]    = get_jump       ("fw_DbgSi_LoadDmx       ", fw.fw_process_tot[i],                    0x10, 0x30, base, result);
    }
    fw.fw_utc_time[i]             = search_variable("fw_utc_time            ", pattern_process_tdt,                     0x6a, 0x6e, 0x70, base, 0x00000000, 0x00250000, result);
    if (fw.fw_utc_time[i] < 0x80000000) {
        //use -0x7B40($gp) for some older f/w
        fw.fw_utc_time[i]         = search_gpvar   ("fw_utc_time            ", pattern_fw_pr_gmt,                fw.fw_gp[i], 0x1a, 0x40, base, 0x00000000, 0x00250000, result);
    }
    fw.fw_utctolocal_time[i]      = search_function("fw_utctolocal_time     ", pattern_fw_utctolocal_time,                          0x40, base, 0x00000000, 0x00250000, result);

    fw.fw_process_eit[i]          = search_function("fw_process_eit         ", pattern_fw_process_eit,                              0x38, base, 0x00000000, 0x00250000, result);
    fw.TAP_GetCurrentEvent[i]     = search_function("TAP_GetCurrentEvent    ", pattern_TAP_GetCurrentEvent,                         0xa0, base, 0x00000000, 0x00250000, result);
    fw.fw_event_table[i]          = search_variable("fw_event_table         ", pattern_TAP_GetCurrentEvent,             0x7e, 0x82, 0xa0, base, 0x00000000, 0x00250000, result);
    fw.fw_event_table_size[i]     = search_variable("fw_event_table_size    ", pattern_TAP_GetCurrentEvent,             0x9e, 0x9e, 0xa0, base, 0x00000000, 0x00250000, result);
    fw.fw_max_events[i]           = search_variable("fw_max_events          ", pattern_TAP_GetCurrentEvent,             0x86, 0x86, 0xa0, base, 0x00000000, 0x00250000, result);
    fw.fw_event_alloc[i]          = search_function("fw_event_alloc         ", pattern_fw_event_alloc,                              0x40, base, 0x00000000, 0x00250000, result);
    fw.fw_event_buffer[i]         = search_variable("fw_event_buffer        ", pattern_fw_event_alloc,                  0x32, 0x36, 0x40, base, 0x00000000, 0x00250000, result);
    fw.fw_event_delete[i]         = search_function("fw_event_delete        ", pattern_fw_event_delete,                             0x40, base, 0x00000000, 0x00250000, result);
    if (fw.fw_event_delete[i] == 0) {
    fw.fw_event_delete[i]         = search_function("fw_event_delete        ", pattern_fw_event_delete_1288,                        0x40, base, 0x00000000, 0x00250000, result);
    }
    fw.fw_process_dsmcc[i]        = search_function("fw_process_dsmcc       ", pattern_fw_process_dsmcc,                            0x40, base, 0x00000000, 0x00250000, result);
    if (fw.fw_process_dsmcc[i] == 0) {
    fw.fw_process_dsmcc[i]        = search_function("fw_process_dsmcc       ", pattern_fw_process_dsmcc2,                           0x5c, base, 0x00000000, 0x00250000, result);
    }
    if (fw.fw_process_dsmcc[i] == 0) {
    fw.fw_process_dsmcc[i]        = search_function("fw_process_dsmcc       ", pattern_fw_process_dsmcc3,                           0x58, base, 0x00000000, 0x00250000, result);
    }

    fw.fw_mheg_entry[i]           = search_gpvar   ("fw_mheg_entry          ", pattern_fw_mheg_entry,            fw.fw_gp[i], 0x16, 0x40, base, 0x00000000, 0x00250000, result);
    if (fw.fw_mheg_entry[i] == 0) {
    fw.fw_mheg_entry[i]           = search_gpvar   ("fw_mheg_entry          ", pattern_fw_mheg_entry2,           fw.fw_gp[i], 0x16, 0x40, base, 0x00000000, 0x00250000, result);
    }
    fw.fw_hdd_move[i]             = search_function("fw_hdd_move            ", pattern_fw_hdd_move,                                 0x50, base, 0x00000000, 0x00250000, result);
    fw.fw_key_map_patch[i]        = search_function("fw_key_map_patch       ", pattern_fw_key_map_patch,                            0x40, base, 0x00000000, 0x00250000, result);
    if (fw.fw_key_map_patch[i] != 0) {
        printf("\nsearching for %s ... ", "fw_key_map_patch2      ");
        fw.fw_key_map_patch2[i]   = fw.fw_key_map_patch[i] + 0x58;
        if (fw.fw_key_map_patch2[i] == 0) {
            printf("not found");
        }
        else {
            printf("found at %08x", fw.fw_key_map_patch2[i]);
        }
    }
//    fw.fw_led_patch[i]          = search_function("fw_led_patch           ", pattern_fw_led_patch,                                0x40, base, 0x00000000, 0x00250000, result);
    fw.fw_display_func[i]         = search_function("fw_display_func        ", pattern_fw_display_func,                             0x40, base, 0x00000000, 0x00250000, result);
    fw.fw_display_string[i]       = search_gpvar   ("fw_display_string      ", pattern_fw_translate_and_display, fw.fw_gp[i], 0x2a, 0x40, base, 0x00000000, 0x00250000, result);

    fw.fw_ffindfirst[i]           = search_gpvar   ("fw_tap_load_data0      ", pattern_fw_ffindfirst,            fw.fw_gp[i], 0x22, 0x40, base, 0x00000000, 0x00250000, result);
    fw.fw_tap_load[i]             = search_function("fw_tap_load            ", pattern_fw_tap_load,                                 0x40, base, 0x00000000, 0x00250000, result);
    fw.fw_tap_load_data0[i]       = fw.fw_ffindfirst[i];
//    fw.fw_tap_load_data1[i]       = search_gpvar   ("fw_tap_load_data1      ", pattern_fw_tap_load_data1,        fw.fw_gp[i], 0x36, 0x40, base, 0x00000000, 0x00250000, result);
    fw.fw_tap_load_data2[i]       = search_gpvar   ("fw_tap_load_data2      ", pattern_fw_tap_load_data1,        fw.fw_gp[i], 0x6a, 0x78, base, 0x00000000, 0x00250000, result);

    fw.fw_num_ext_functions[i]    = search_variable("fw_num_ext_functions   ", pattern_fw_TAP_GetSystemProc,            0x3a, 0x3a, 0x50, base, 0x00000000, 0x00250000, result);
    fw.fw_ext_function_table[i]   = search_variable("fw_ext_function_table  ", pattern_fw_TAP_GetSystemProc,            0x42, 0x4e, 0x50, base, 0x00000000, 0x00250000, result);

    fw.fw_enable_tshift[i]        = search_function("fw_enable_tshift       ", pattern_fw_enable_tshift,                            0x40, base, 0x00000000, 0x00250000, result);
    fw.fw_stop_tshift[i]          = search_function("fw_stop_tshift         ", pattern_fw_stop_tshift,                              0x40, base, 0x00000000, 0x00250000, result);
    if (fw.fw_stop_tshift[i] == 0) {
    fw.fw_stop_tshift[i]          = search_function("fw_stop_tshift         ", pattern_fw_stop_tshift2,                             0x40, base, 0x00000000, 0x00250000, result);
    }
    fw.fw_mheg_disable[i]         = search_function("fw_mheg_disable        ", pattern_fw_mheg_disable,                             0x3c, base, 0x00000000, 0x00250000, result);
    fw.fw_ioctl[i]                = search_function("fw_ioctl               ", pattern_fw_ioctl,                                    0x40, base, 0x00000000, 0x00250000, result);
    fw.fw_epgstck_patch[i]        = search_function("fw_epgstck_patch       ", pattern_fw_epgstck_patch,                            0x40, base, 0x00000000, 0x00250000, result);

    fw.fw_channel_scale[i]        = search_function("fw_channel_scale       ", pattern_fw_channel_scale,                            0x40, base, 0x00000000, 0x00250000, result);
    if (fw.fw_channel_scale[i] == 0) {
    fw.fw_channel_scale[i]        = search_function("fw_channel_scale       ", pattern_fw_channel_scale2,                           0x40, base, 0x00000000, 0x00250000, result);
    }
    fw.fw_set_wss[i]              = search_function("fw_set_wss             ", pattern_fw_set_wss,                                  0x40, base, 0x00000000, 0x00250000, result);
    if (fw.fw_set_wss[i] == 0) {
    fw.fw_set_wss[i]              = search_function("fw_set_wss             ", pattern_fw_set_wss2,                                 0x40, base, 0x00000000, 0x00250000, result);
    }
    fw.fw_etc_structure[i]        = search_gpvar   ("fw_etc_structure       ", pattern_TAP_GetSignalLevel,       fw.fw_gp[i], 0x4a, 0x50, base, 0x00000000, 0x00250000, result);
    fw.fw_afd[i]                  = search_gpvar   ("fw_afd                 ", pattern_fw_afd,                   fw.fw_gp[i], 0x3e, 0x40, base, 0x00000000, 0x00250000, result);
    fw.fw_sequence_header[i]      = search_gpvar   ("fw_sequence_header     ", pattern_fw_sequence_header,       fw.fw_gp[i], 0x22, 0x40, base, 0x00000000, 0x00250000, result);
    fw.fw_get_genre_str[i]        = search_function("fw_get_genre_str       ", pattern_fw_get_genre_str,                            0x40, base, 0x00000000, 0x00250000, result);
    fw.fw_epg_highlighted_event[i]= search_variable("fw_epg_highlighted_event", pattern_fw_epg_highlighted_event,       0x26, 0x2e, 0x40, base, 0x00000000, 0x00250000, result);
    fw.fw_epg_refresh[i]          = search_gpvar   ("fw_epg_refresh         ", pattern_fw_epg_refresh,           fw.fw_gp[i], 0x1a, 0x30, base, 0x00000000, 0x00250000, result);
    return 0;
}



void output_filename(FILE *OpC)
{
    int i;

    fprintf(OpC,"\n//  sysid  version  filename");
    for (i=0; i<num_firmwares; i++) {
        fprintf(OpC,"\n// %6d  %7x  %s", fw.fw_sysid[i], fw.fw_appl_version[i], fw_filename[i]);
    }
}

void output_type(FILE *OpH, FILE *OpC, char *type_str, dword *addr, char *comment)
{
    int i;

    fprintf(OpH,"\nextern type_firmware %s[];", type_str);

    fprintf(OpC,"\n");
    if (comment) fprintf(OpC,"\n// %s", comment);
    fprintf(OpC,"\ntype_firmware %s[] = { ", type_str);
    for (i=0; i<num_firmwares; i++) {
        fprintf(OpC,"\n    {%5d, 0x%04x, 0x%08x},", fw.fw_sysid[i], fw.fw_appl_version[i], addr[i]);
    }
    fprintf(OpC,"\n    {0, 0, 0},");
    fprintf(OpC,"\n};");
}


int main(int argc, char **argv)
{
    FILE *Ip;
    int fileno;
    int flen;
    byte *fbuf;
    int i;
    int ret;
    FILE *OpH, *OpC;
    SYSTEMTIME SystemTime;
    struct _finddata_t raw_file;
    long hFile;
    char path[0x40] = ".\\";
    char fname[0x100] = "firmware_sys_456";
    char ext[0x10] = "raw";
    char filename[0x100];
    int numin;
    int sysid;
    char *str_ptr;
    int got_filename;


     got_filename = 0;
     for(i=1; i<argc; i++) {
        if ((argv[i][0] == '-') || (argv[i][0] == '/')) {
            switch (argv[i][1]) {
                case 'p' :
                    sscanf(argv[++i],"%s", path);
                    break;
                case 'f' :
                    sscanf(argv[++i],"%s", fname);
                    got_filename = 1;
                    break;
                case 'x' :
                    sscanf(argv[++i],"%s", ext);
                    break;
                case 'h' :
                case '?' :
                default  :
                    printf("\nusage:");
                    printf("\n    generate_patches <-p path> <-f fname> <-x ext>");
                    return 1;
            }
        }
    }

    i = 0;
    if (got_filename) {
        strcpy(filename, path);
        strcat(filename, "\\");
        strcat(filename, fname);
        strcat(filename, ".");
        strcat(filename, ext);

        //scan it for sysid-ddd-
        numin = 0;
        str_ptr = strstr(filename, "Sys");
        if (!str_ptr) str_ptr = strstr(filename, "sys");
        if (!str_ptr) str_ptr = strstr(filename, "SYS");
        if (str_ptr) {
            str_ptr+=4;
            numin = sscanf(str_ptr, "%d", &sysid);
        }
        if (numin != 1) {
            printf("\n%d: cannot detect sysid in file: %s", numin, filename);
        }
        else {
            printf("\nfound sysid %d in file: %s", sysid, filename);
            fw.fw_sysid[i] = sysid;
            strcpy(fw_filename[i], fname);
            strcat(fw_filename[i], ".");
            strcat(fw_filename[i], ext);
            printf("\nReading %s", filename);
            Ip = fopen(filename, "rb");
            if (!Ip) {
                printf("\ncannot open %s", filename);
                return -1;
            }
            fileno = _fileno(Ip);
            flen = _filelength(fileno);
            fbuf = (UCHAR *)malloc(flen);
            if (!fbuf) printf("\nfbuf=%x",fbuf);
            fread(fbuf, flen, 1, Ip);
            fclose(Ip);

            test(fbuf, i);
            i++;
        }
    }
    else {
        //get all the filenames called .raw
        strcpy(filename, path);
        strcat(filename, "\\");
        strcat(filename, "*.");
        strcat(filename, ext);
        if ((hFile = _findfirst(filename, &raw_file )) == -1) {
            printf( "\nNo *.%s files in current directory, %s", ext, filename);
            return -1;
        }
        else {
            // Find the .raw files
            do {
                printf( "\nfound:  %-12s, %9ld", raw_file.name, raw_file.size);

                //scan it for sysid-ddd-
                numin = 0;
                str_ptr = strstr(raw_file.name, "Sys");
                if (!str_ptr) str_ptr = strstr(raw_file.name, "sys");
                if (!str_ptr) str_ptr = strstr(raw_file.name, "SYS");
                if (str_ptr) {
                    str_ptr+=4;
                    numin = sscanf(str_ptr, "%d", &sysid);
                }
                if (numin != 1) {
                    printf("\n%d: cannot detect sysid in file: %s", numin, raw_file.name);
                }
                else {
                    printf("\nfound sysid %d in file: %s", sysid, raw_file.name);
                    fw.fw_sysid[i] = sysid;
                    strcpy(fw_filename[i], raw_file.name);
                    strcpy(filename, path);
                    strcat(filename, "\\");
                    strcat(filename, raw_file.name);
                    printf("\nReading %s", filename);
                    Ip = fopen(filename, "rb");
                    if (!Ip) {
                        printf("\ncannot open %s", filename);
                        return -1;
                    }
                    fileno = _fileno(Ip);
                    flen = _filelength(fileno);
                    fbuf = (UCHAR *)malloc(flen);
                    if (!fbuf) printf("\nfbuf=%x",fbuf);
                    fread(fbuf, flen, 1, Ip);
                    fclose(Ip);

                    test(fbuf, i);
                    i++;
                }

                ret = _findnext(hFile, &raw_file);
            } while (ret == 0);

            _findclose(hFile);
        }
    }
    num_firmwares = i;


    OpH = fopen("patches.h","wt");
    if (!OpH) return 1;
    OpC = fopen("patches.c","wt");
    if (!OpC) {
        fclose(OpH);
        return 1;
    }
    GetLocalTime(&SystemTime);
    fprintf(OpH,"//");
    fprintf(OpH,"\n//autogenerated file on %02d/%02d/%04d", SystemTime.wDay, SystemTime.wMonth, SystemTime.wYear);
    fprintf(OpH,"\n//");
    fprintf(OpH,"\n#ifndef _PATCHES_");
    fprintf(OpH,"\n#define _PATCHES_");
    fprintf(OpH,"\n");
    fprintf(OpC,"//");
    fprintf(OpC,"\n//autogenerated file on %02d/%02d/%04d from:", SystemTime.wDay, SystemTime.wMonth, SystemTime.wYear);
    fprintf(OpC,"\n//");
    output_filename(OpC);
    fprintf(OpC,"\n//");
    fprintf(OpC,"\n#include <tap.h>");
    fprintf(OpC,"\n#include \"exTap.h\"");
    fprintf(OpC,"\n#include \"patches.h\"");


    output_type(OpH, OpC, "fwlist_tap_table",            fw.fw_tap_table,            "the tap table base address; 0x20 bytes per tap");
    output_type(OpH, OpC, "fwlist_tap_ident",            fw.fw_tap_ident,            "the current tap id");
    output_type(OpH, OpC, "fwlist_gp",                   fw.fw_gp,                   "comment");
    output_type(OpH, OpC, "fwlist_os_free",              fw.fw_os_free,              "the address of the opererating system free() function");
    output_type(OpH, OpC, "fwlist_add_task_queue",       fw.fw_add_task_queue,       "the address of an add task function");
    output_type(OpH, OpC, "fwlist_shutdown_irq_disable", fw.fw_shutdown_irq_disable, "comment");
    output_type(OpH, OpC, "fwlist_shutdown_irq_enable",  fw.fw_shutdown_irq_enable,  "comment");
    output_type(OpH, OpC, "fwlist_shutdown_task",        fw.fw_shutdown_task,        "comment");
    output_type(OpH, OpC, "fwlist_process_tdt",          fw.fw_process_tdt,          "the address of the process_tdt() function");
    output_type(OpH, OpC, "fwlist_process_tot",          fw.fw_process_tot,          "the address of the process_tot() function");
    output_type(OpH, OpC, "fwlist_DbgSi_LoadDmx",        fw.fw_DbgSi_LoadDmx,        "the address of the DbgSi_LoadDmx() function");
    output_type(OpH, OpC, "fwlist_utc_time",             fw.fw_utc_time,             "the address of the current utc time");
    output_type(OpH, OpC, "fwlist_utctolocal_time",      fw.fw_utctolocal_time,      "the address of the utc2local time function");

    output_type(OpH, OpC, "fwlist_process_eit",          fw.fw_process_eit,          "comment");
    output_type(OpH, OpC, "fwlist_TAP_GetCurrentEvent",  fw.TAP_GetCurrentEvent,     "comment");
    output_type(OpH, OpC, "fwlist_event_table",          fw.fw_event_table,          "the base of the internal event table");
    output_type(OpH, OpC, "fwlist_event_table_size",     fw.fw_event_table_size,     "the size of the internal event table, 0x40/0x44 bytes per event");
    output_type(OpH, OpC, "fwlist_max_events",           fw.fw_max_events,           "comment");
    output_type(OpH, OpC, "fwlist_event_alloc",          fw.fw_event_alloc,          "the address of the event buffer allocation function");
    output_type(OpH, OpC, "fwlist_event_buffer",         fw.fw_event_buffer,         "the address of the event buffer structure");
    output_type(OpH, OpC, "fwlist_event_delete",         fw.fw_event_delete,         "the address of the delete event function");
    output_type(OpH, OpC, "fwlist_process_dsmcc",        fw.fw_process_dsmcc,        "comment");
    output_type(OpH, OpC, "fwlist_mheg_entry",           fw.fw_mheg_entry,           "the address of the mheg entry structure, [1225: $gp-0x74f4]");
    output_type(OpH, OpC, "fwlist_hdd_move",             fw.fw_hdd_move,             "the address of the internal hdd_move function");
    output_type(OpH, OpC, "fwlist_key_map_patch",        fw.fw_key_map_patch,        "comment");
    output_type(OpH, OpC, "fwlist_key_map_patch2",       fw.fw_key_map_patch2,       "comment");
    output_type(OpH, OpC, "fwlist_display_func",         fw.fw_display_func,         "the address of the front panel display function");
    output_type(OpH, OpC, "fwlist_display_string",       fw.fw_display_string,       "the location of the firmware front panel current display string [1225: $gp-0x5fe0]");
    output_type(OpH, OpC, "fwlist_tap_load",             fw.fw_tap_load,             "the address of the tap load function");
    output_type(OpH, OpC, "fwlist_tap_load_data0",       fw.fw_tap_load_data0,       "the address of the tap load data0 [1225: $gp-0x6194]");
//    output_type(OpH, OpC, "fwlist_tap_load_data1",       fw.fw_tap_load_data1,       "the address of the tap load data1 [1225: $gp-0x7af0]");
    output_type(OpH, OpC, "fwlist_tap_load_data2",       fw.fw_tap_load_data2,       "the address of the tap load data2 [1225: $gp-0x7b00]");
    output_type(OpH, OpC, "fwlist_num_ext_functions",    fw.fw_num_ext_functions,    "comment");
    output_type(OpH, OpC, "fwlist_ext_function_table",   fw.fw_ext_function_table,   "comment");

    output_type(OpH, OpC, "fwlist_enable_tshift",        fw.fw_enable_tshift,         "the address of the enable_tshift function");
    output_type(OpH, OpC, "fwlist_stop_tshift",          fw.fw_stop_tshift,           "the address of the stop_tshift function");
    output_type(OpH, OpC, "fwlist_mheg_disable",         fw.fw_mheg_disable,          "the address of the mheg_disable function");
    output_type(OpH, OpC, "fwlist_ioctl",                fw.fw_ioctl,                 "the address of the ioctl function");
    output_type(OpH, OpC, "fwlist_epgstck_patch",        fw.fw_epgstck_patch,         "the address for the epgstick patch");
    output_type(OpH, OpC, "fwlist_channel_scale",        fw.fw_channel_scale,         "the address for the channel_scale() function");
    output_type(OpH, OpC, "fwlist_set_wss",              fw.fw_set_wss,               "the address for the set_wss() function");

    output_type(OpH, OpC, "fwlist_etc_structure",        fw.fw_etc_structure,         "the address of the eeprom etc structure");
    output_type(OpH, OpC, "fwlist_afd",                  fw.fw_afd,                   "the address of the afd variable");
    output_type(OpH, OpC, "fwlist_sequence_header",      fw.fw_sequence_header,       "the address of the mpeg sequence_header buffer");
    output_type(OpH, OpC, "fwlist_get_genre_str",        fw.fw_get_genre_str,         "the address of the get_genre_str function");
    output_type(OpH, OpC, "fwlist_epg_highlighted_event",fw.fw_epg_highlighted_event, "the address of the epg highlighted event");
    output_type(OpH, OpC, "fwlist_epg_refresh",          fw.fw_epg_refresh,           "the address of the epg refresh variable");

    fprintf(OpH,"\n");
    fprintf(OpH,"\n#endif //_PATCHES_");
    fprintf(OpH,"\n");
    fprintf(OpC,"\n");
    fclose(OpH);
    fclose(OpC);

    return 0;
}

#endif
